from __future__ import annotations

from typing import TYPE_CHECKING, NamedTuple

if TYPE_CHECKING:
    from rattr.analyser.types import ImportIrs
    from rattr.models.ir import FileIr, FunctionIr
    from rattr.models.symbol import Call, Class, Func


class IrTarget(NamedTuple):
    symbol: Func | Class
    ir: FunctionIr


class IrCall(NamedTuple):
    caller: Func | Class
    symbol: Call


class IrEnvironment(NamedTuple):
    target_ir: FileIr
    import_irs: ImportIrs


class IrCallTreeNode(NamedTuple):
    target: IrTarget

    edge_in: IrCall | None
    edges_out: list[IrCall]

    children: list[IrCallTreeNode]

    @classmethod
    def new(cls, target: IrTarget, call: IrCall | None) -> IrCallTreeNode:
        return cls(
            target=IrTarget(symbol=target.symbol, ir=target.ir.copy()),
            edge_in=call,
            edges_out=[
                IrCall(caller=target.symbol, symbol=symbol)
                for symbol in sorted(target.ir["calls"], key=lambda c: c.id)
            ],
            children=[],
        )



from __future__ import annotations

import copy
from typing import TYPE_CHECKING

import pytest

from rattr.models.context import SymbolTable
from rattr.models.symbol import CallInterface, Func, Name

if TYPE_CHECKING:
    from tests.shared import MakeSymbolTableFn


@pytest.fixture
def symbol_table(make_symbol_table: MakeSymbolTableFn) -> SymbolTable:
    return make_symbol_table(
        {
            "var": Name("var"),
            "x": Name("x"),
            "y": Name("y"),
            "z": Name("z"),
            "my_func": Func("my_func", interface=CallInterface(args=["arg"])),
        }
    )


class TestSymbolTable:
    def test_names(self, symbol_table: SymbolTable):
        assert symbol_table.names == {"var", "x", "y", "z", "my_func"}

    def test_symbols(self, symbol_table: SymbolTable):
        assert list(symbol_table.symbols) == [
            Name("var"),
            Name("x"),
            Name("y"),
            Name("z"),
            Func("my_func", interface=CallInterface(args=["arg"])),
        ]

    def test_get(self, symbol_table: SymbolTable):
        assert symbol_table.get("x") == Name("x")

    def test_get_absent(self, symbol_table: SymbolTable):
        assert symbol_table.get("nope") is None
        assert symbol_table.get("nope", None) is None


class TestSymbolTableAdd:
    def test_add(self):
        symbol_table = SymbolTable()
        symbol = Name("new_symbol")
        assert symbol not in symbol_table

        symbol_table.add(symbol)
        assert symbol.id in symbol_table
        assert list(symbol_table.symbols) == [symbol]

    def test_add_multiple(self):
        symbol_table = SymbolTable()
        symbols = [Name("new_symbol"), Name("AnotherNewSymbol")]

        for symbol in symbols:
            assert symbol not in symbol_table

        symbol_table.add(symbols)

        for symbol in symbols:
            assert symbol.id in symbol_table

        assert list(symbol_table.symbols) == symbols

    def test_add_duplicate(self, symbol_table: SymbolTable):
        before = copy.deepcopy(symbol_table)
        after = copy.deepcopy(symbol_table)

        after.add(Name("x"))
        assert before == after


class TestSymbolTableRemove:
    def test_remove_by_id(self, symbol_table: SymbolTable):
        id = "x"
        assert id in symbol_table

        symbol_table.remove(id)
        assert id not in symbol_table

    def test_remove_by_symbol(self, symbol_table: SymbolTable):
        symbol = symbol_table["x"]
        assert symbol.id in symbol_table
        assert symbol in symbol_table.symbols

        symbol_table.remove(symbol)
        assert symbol.id not in symbol_table
        assert symbol not in symbol_table.symbols

    def test_remove_multiple(self, symbol_table: SymbolTable):
        ids = ["x", "y", "z"]

        for id in ids:
            assert id in symbol_table

        symbol_table.remove(ids)

        for id in ids:
            assert id not in symbol_table

    def test_remove_missing(self, symbol_table: SymbolTable):
        before = copy.deepcopy(symbol_table)
        after = copy.deepcopy(symbol_table)

        with pytest.raises(KeyError):
            after.remove("not_a_real_symbol")
        assert before == after


class TestSymbolTablePop:
    def test_pop_by_id(self, symbol_table: SymbolTable):
        id = "x"
        assert id in symbol_table

        symbol = symbol_table.pop(id)
        assert symbol.id == id
        assert id not in symbol_table

    def test_pop_by_symbol(self, symbol_table: SymbolTable):
        symbol = Name("x")
        assert symbol in symbol_table.symbols

        symbol_ = symbol_table.pop(symbol)
        assert symbol == symbol_
        assert symbol not in symbol_table.symbols

    def test_pop_missing(self, symbol_table: SymbolTable):
        before = copy.deepcopy(symbol_table)
        after = copy.deepcopy(symbol_table)

        symbol = after.pop("not_a_real_symbol")
        assert symbol is None
        assert before == after



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from rattr.models.ir import FileIr
from rattr.models.results import FileResults, FunctionResults

if TYPE_CHECKING:
    from typing import Any


here = Path(__file__).resolve().parent

code_dir = here / "code"
results_dir = here / "results"
irs_dir = here / "irs"

code_files = [f for f in code_dir.rglob("*.py")]
results_files = [
    (results_dir / code_file.relative_to(code_dir)).with_suffix(".json")
    for code_file in code_files
]
ir_files = [
    (irs_dir / code_file.relative_to(code_dir)).with_suffix(".json")
    for code_file in code_files
]


def assert_actual_and_expected_have_the_same_functions_results(
    actual: FileResults,
    expected: dict[str, Any],
) -> None:
    actual_results_functions = set(actual.keys())
    expected_results_functions = set(expected.keys())

    missing_actual = list(expected_results_functions - actual_results_functions)
    missing_expected = list(actual_results_functions - expected_results_functions)

    assert not missing_actual, f"actual results missing functions: {missing_actual}"
    assert (
        not missing_expected
    ), f"expected results missing functions: {missing_expected}"


def assert_actual_and_expected_have_the_same_functions_irs(
    actual: FileIr,
    expected: FileIr,
) -> None:
    actual_results_functions = set(actual.keys())
    expected_results_functions = set(expected.keys())

    missing_actual = list(expected_results_functions - actual_results_functions)
    missing_expected = list(actual_results_functions - expected_results_functions)

    assert not missing_actual, f"actual results missing functions: {missing_actual}"
    assert (
        not missing_expected
    ), f"expected results missing functions: {missing_expected}"


def results_diff(got: set[str], expected: set[str]) -> list[str]:
    diff: set[str] = set()

    for missing in expected - got:
        diff.add(f"- {missing}")

    for missing in got - expected:
        diff.add(f"+ {missing}")

    return list(diff)


def results_function_diffs(
    fn_results: FunctionResults,
    expected_fn_results: dict[str, list[str]],
) -> dict[str, list[str]]:
    return {
        k: results_diff(fn_results[k], set(expected_fn_results[k]))
        for k in ("sets", "gets", "dels", "calls")
    }



from __future__ import annotations

import os
import sys
from functools import cache
from pathlib import Path
from typing import TYPE_CHECKING

from rattr.module_locator.exc import RattrSysPathNotPopulated

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Final

    from rattr.module_locator.types import ModuleName


rattr_root: Final = Path(__file__).parents[2]


@cache
def locate_module_in_python_path(modulename: ModuleName) -> list[Path]:
    """Return the origins of the given module in the Python path."""
    return [
        module_location
        for search_location in iter_python_path_dirs()
        if (module_location := find_module_in_path(search_location, modulename))
    ]


def find_module_in_path(
    python_path: Path,
    modulename: ModuleName,
) -> Path | None:
    """Return the module's definition file in the given path, if extant."""
    # NOTE
    # Handle explicitly because in PyPy the empty module will match the always present
    # file: .../pypy3.X/__init__.py
    if modulename == "":
        return None

    module_parts = modulename.split(".")
    install_location = python_path.resolve()

    for part in module_parts:
        install_location /= part

    if install_location.is_dir():
        install_location /= "__init__.py"
    else:
        install_location = install_location.with_suffix(".py")

    if not install_location.exists():
        return None

    return install_location


def iter_python_path_dirs() -> Iterator[Path]:
    """Yield the extant dirs in the Python path.

    In practice this is usually the interpreter's CWD and the $PYTHONPATH, but retrieved
    from the interpreter itself not from the environment vars (which may be
    unpopulated), this is done using `sys.path` following the logic from the docs [1].

        [1] - https://docs.python.org/3/library/sys.html?highlight=sys#sys.path

    Raises:
        RattrSysPathNotPopulated: sys.paths is not populated.

    Yields:
        Iterator[Path]: The dirs in the Python path.
    """
    if not sys.path:
        raise RattrSysPathNotPopulated

    # When the sys.paths[0] is the empty string then the interpreter uses the current
    # directory (see `sys.path` docs)
    if sys.path[0] == "":
        first = os.getcwd()
    else:
        first = sys.path[0]

    yield from (
        python_path_dir
        for python_path_dirname in (first, rattr_root, *sys.path[1:])
        if (python_path_dir := Path(python_path_dirname))
        if python_path_dir.exists()
    )



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from rattr import _version
from rattr.cli._util import multi_paragraph_wrap
from rattr.config import Output

if TYPE_CHECKING:
    from rattr.cli._argparse import ArgumentParser


def add_common_arguments(parser: ArgumentParser) -> ArgumentParser:
    """Apply the arguments common to the cli and toml."""
    parser = add_follow_imports_argument(parser)
    parser = add_exclude_imports_argument(parser)
    parser = add_exclude_names_argument(parser)
    parser = add_warning_level_argument(parser)
    parser = add_format_path_arguments(parser)
    parser = add_permissiveness_arguments(parser)
    parser = add_stdout_arguments(parser)

    return parser


def add_version_argument(parser: ArgumentParser) -> ArgumentParser:
    version_group = parser.add_argument_group()
    version_group.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {_version.version}",
    )

    return parser


def add_toml_config_override_argument(parser: ArgumentParser) -> ArgumentParser:
    parser.add_argument(
        "-c",
        "--config",
        default=None,
        type=Path,
        required=False,
        help=multi_paragraph_wrap(
            """\
            override the default 'pyproject.toml' with another config file
            """
        ),
        metavar="TOML",
        dest="pyproject_toml_override",
    )

    return parser


def add_follow_imports_argument(parser: ArgumentParser) -> ArgumentParser:
    follow_imports_group = parser.add_argument_group()
    follow_imports_group.add_argument(
        "-f",
        "--follow-imports",
        default=1,
        type=int,
        choices=[0, 1, 2, 3],
        help=multi_paragraph_wrap(
            """\
            >follow imports level meanings:
            >    0 - do not follow imports
            >    1 - follow imports to local modules \033[1m(default)\033[0m
            >    2 - follow imports to local and pip installed modules
            >    3 - follow imports to local, pip installed, and stdlib modules

            >NB: following stdlib imports when using CPython will cause issues

            >TOML example: follow-imports=1
            """
        ),
        dest="_follow_imports_level",
    )

    return parser


def add_exclude_imports_argument(parser: ArgumentParser) -> ArgumentParser:
    exclude_imports_group = parser.add_argument_group()
    exclude_imports_group.add_argument(
        "-F",
        "--exclude-import",
        action="append",
        type=str,
        help=multi_paragraph_wrap(
            """\
            >do not follow imports to modules matching the given pattern,
            >regardless of the level of \033[1m-f\033[0m

            >TOML example: exclude-imports=['a', 'b']
            """
        ),
        metavar="PATTERN",
        dest="_excluded_imports",
    )

    return parser


def add_exclude_names_argument(parser: ArgumentParser) -> ArgumentParser:
    exclude_names_group = parser.add_argument_group()
    exclude_names_group.add_argument(
        "-x",
        "--exclude",
        action="append",
        type=str,
        help=multi_paragraph_wrap(
            """\
            >exclude functions and classes matching the
            >given regular expression from being analysed

            >TOML example: exclude=['a', 'b']
            """
        ),
        metavar="PATTERN",
        dest="_excluded_names",
    )

    return parser


def add_warning_level_argument(parser: ArgumentParser) -> ArgumentParser:
    warning_level_group = parser.add_argument_group()
    warning_level_group.add_argument(
        "-w",
        "--warning-level",
        default="default",
        type=str,
        choices=["none", "local", "default", "all"],
        help=multi_paragraph_wrap(
            """\
            >warnings level meaning:
            >    none    - do not show warnings
            >    local   - show warnings for <file>
            >    default - show warnings for all files \033[1m(default)\033[0m
            >    all     - show warnings for all files, including low-priority

            >NB: errors and fatal errors are always shown

            >TOML example: warning-level='all'
            """
        ),
        dest="_warning_level",
    )

    return parser


def add_format_path_arguments(parser: ArgumentParser) -> ArgumentParser:
    show_path_group = parser.add_argument_group()

    show_path_group.add_argument(
        "-H",
        "--collapse-home",
        action="store_true",
        help=multi_paragraph_wrap(
            """\
            >collapse the user's home directory in error messages
            >E.g.: "/home/user/path/to/file" becomes "~/path/to/file"

            >TOML example: collapse-home=true
            """
        ),
        dest="collapse_home",
    )
    show_path_group.add_argument(
        "-T",
        "--truncate-deep-paths",
        action="store_true",
        help=multi_paragraph_wrap(
            """\
            >truncate deep file paths in error messages
            >E.g.: "/home/user/very/deep/dir/to/file" becomes "/.../dir/to/file"

            >TOML example: truncate-deep-paths=true
            """
        ),
        dest="truncate_deep_paths",
    )

    return parser


def add_permissiveness_arguments(parser: ArgumentParser) -> ArgumentParser:
    strict_or_permissive_group = parser.add_argument_group()
    strict_or_permissive_mutex_group = (
        strict_or_permissive_group.add_mutually_exclusive_group()
    )

    strict_or_permissive_mutex_group.add_argument(
        "--strict",
        action="store_true",
        help=multi_paragraph_wrap(
            """\
            >select strict mode, i.e. fail on any error

            >TOML example: strict=true
            """
        ),
        dest="is_strict",
    )
    strict_or_permissive_mutex_group.add_argument(
        "--threshold",
        default=0,
        type=int,
        help=multi_paragraph_wrap(
            """\
            set the 'badness' threshold, where 0 is infinite
            \033[1m(default: --threshold 0)\033[0m

            >typical badness values:
            >    +0 - info
            >    +1 - warning
            >    +5 - error
            >    +∞ - fatal

            >NB: badness is calculated form the target file and simplification stage

            >TOML example: threshold=10
            """
        ),
        metavar="N",
        dest="threshold",
    )

    return parser


def add_stdout_arguments(parser: ArgumentParser) -> ArgumentParser:
    stdout_group = parser.add_argument_group()
    stdout_group.add_argument(
        "-o",
        "--stdout",
        default=Output.results,
        type=Output,
        choices=list(Output),
        help=multi_paragraph_wrap(
            """\
            >output selection:
            >    silent  - do not print to stdout
            >    ir      - print the intermediate representation to stdout
            >    results - print the results to stdout \033[1m(default)\033[0m

            >TOML example: stdout='results'
            """
        ),
        dest="stdout",
    )

    return parser


def add_target_file_argument(parser: ArgumentParser) -> ArgumentParser:
    target_file_group = parser.add_argument_group()
    target_file_group.add_argument(
        "target",
        type=Path,
        help=multi_paragraph_wrap(
            """\
            >the target source file
            """
        ),
        metavar="<file>",
    )

    return parser



from __future__ import annotations

from rattr.models.util._types import OutputIrs
from rattr.models.util.serialise import (
    deserialise,
    serialise,
    serialise_irs,
)

__all__ = [
    "OutputIrs",
    "deserialise",
    "serialise",
    "serialise_irs",
]



from __future__ import annotations

import copy
from typing import TYPE_CHECKING

import pytest

from rattr.models.ir import FileIr, FunctionIr
from rattr.models.symbol import CallInterface, Func, Name, UserDefinedCallableSymbol

if TYPE_CHECKING:
    from rattr.models.context import Context
    from tests.shared import MakeRootContextFn


@pytest.fixture
def context_a(
    make_root_context: MakeRootContextFn,
) -> Context:
    return make_root_context([Name("hi")], include_root_symbols=True)


@pytest.fixture
def context_b(
    make_root_context: MakeRootContextFn,
) -> Context:
    return make_root_context([Name("not_hi")], include_root_symbols=True)


@pytest.fixture
def underlying_file_ir_a() -> dict[UserDefinedCallableSymbol, FunctionIr]:
    return {
        Func("fn", interface=CallInterface(args=("a",))): {
            "gets": {Name("thing.attr")},
            "sets": set(),
            "dels": set(),
            "calls": set(),
        }
    }


@pytest.fixture
def underlying_file_ir_b() -> dict[UserDefinedCallableSymbol, FunctionIr]:
    # Different attr accessed
    return {
        Func("fn", interface=CallInterface(args=("a",))): {
            "gets": {Name("thing.not_the_other_attr")},
            "sets": set(),
            "dels": set(),
            "calls": set(),
        }
    }


def test_equality_with_matching_context_and_matching_ir(
    context_a: Context,
    underlying_file_ir_a: dict[UserDefinedCallableSymbol, FunctionIr],
):
    lhs = FileIr(
        context=copy.deepcopy(context_a),
        file_ir=copy.deepcopy(underlying_file_ir_a),
    )
    rhs = FileIr(
        context=copy.deepcopy(context_a),
        file_ir=copy.deepcopy(underlying_file_ir_a),
    )
    assert lhs == rhs


def test_equality_with_matching_context_and_non_matching_ir(
    context_a: Context,
    underlying_file_ir_a: dict[UserDefinedCallableSymbol, FunctionIr],
    underlying_file_ir_b: dict[UserDefinedCallableSymbol, FunctionIr],
):
    lhs = FileIr(
        context=copy.deepcopy(context_a),
        file_ir=copy.deepcopy(underlying_file_ir_a),
    )
    rhs = FileIr(
        context=copy.deepcopy(context_a),
        file_ir=copy.deepcopy(underlying_file_ir_b),
    )
    assert lhs != rhs


def test_equality_with_non_matching_context_and_matching_ir(
    context_a: Context,
    context_b: Context,
    underlying_file_ir_a: dict[UserDefinedCallableSymbol, FunctionIr],
):
    lhs = FileIr(
        context=copy.deepcopy(context_a),
        file_ir=copy.deepcopy(underlying_file_ir_a),
    )
    rhs = FileIr(
        context=copy.deepcopy(context_b),
        file_ir=copy.deepcopy(underlying_file_ir_a),
    )
    assert lhs != rhs


def test_equality_with_non_matching_context_and_non_matching_ir(
    context_a: Context,
    context_b: Context,
    underlying_file_ir_a: dict[UserDefinedCallableSymbol, FunctionIr],
    underlying_file_ir_b: dict[UserDefinedCallableSymbol, FunctionIr],
):
    lhs = FileIr(
        context=copy.deepcopy(context_a),
        file_ir=copy.deepcopy(underlying_file_ir_a),
    )
    rhs = FileIr(
        context=copy.deepcopy(context_b),
        file_ir=copy.deepcopy(underlying_file_ir_b),
    )
    assert lhs != rhs



from __future__ import annotations


class RattrUnsupportedError(Exception):
    """Language feature is unsupported by rattr."""

    pass


class RattrUnaryOpInNameable(TypeError):
    """Unary operation found when resolving name."""

    pass


class RattrBinOpInNameable(TypeError):
    """Binary operation found when resolving name."""

    pass


class RattrConstantInNameable(TypeError):
    """Constant found when resolving name."""

    pass


class RattrLiteralInNameable(TypeError):
    """Literal found when resolving name."""

    pass


class RattrComprehensionInNameable(TypeError):
    """Comprehension found when resolving name."""

    pass



from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING

if sys.version_info.major == 3 and sys.version_info.minor >= 11:
    # isort: off
    import tomllib  # type: ignore reportMissingImports
    from tomllib import TOMLDecodeError  # type: ignore reportMissingImports; noqa: F401

    # isort: on
else:
    import tomli as tomllib
    from tomli import TOMLDecodeError  # noqa: F401

if TYPE_CHECKING:
    from typing import Any


# HACK Wrap `tomllib.loads` to handle type hinting on different python versions
def _load_from_file(file: Path) -> dict[str, Any]:
    return tomllib.loads(file.read_text())


def parse_project_toml(
    pyproject_toml: Path | None,
    project_toml_override: Path | None = None,
) -> dict[str, Any]:
    """Return the parsed project toml."""
    if project_toml_override:
        project_toml = project_toml_override
    else:
        project_toml = pyproject_toml

    if project_toml:
        return _load_from_file(project_toml).get("tool", {}).get("rattr", {})

    return {}



from __future__ import annotations

from rattr.cli._util import get_type_name


class TestUtils:
    def test_get_type_name(self):
        assert get_type_name(None) == "NoneType"

        assert get_type_name(True) == "bool"
        assert get_type_name(1) == "int"
        assert get_type_name(1.0) == "float"

        assert get_type_name([1]) == "list[int]"
        assert get_type_name(["1"]) == "list[str]"
        assert get_type_name(["1", 2]) == "list[int | str]"

        assert get_type_name({"1": 1, "2": 2}) == "dict[str, int]"
        assert get_type_name({"1": 1, "2": "two"}) == "dict[str, int | str]"



from __future__ import annotations

import os
from pathlib import Path

from rattr import error
from rattr.config._types import Config
from rattr.config._util import (  # noqa: F401
    find_project_root,
    find_pyproject_toml,
    validate_arguments,
)


def get_current_file() -> Path:
    """Return the current file or error.

    >>> # Config.state.current_file is None
    >>> get_current_file()
    ValueError("unable to get location (current file is None)")

    >>> # Config.state.current_file is Path("my_cool_file.py")
    >>> get_current_file()
    Path("my_cool_file.py")

    Raises:
        ValueError: The current file is not set.

    Returns:
        Path: The file being analysed.
    """
    config = Config()
    file = config.state.current_file

    if file is None:
        raise ValueError("unable to get location (current file is None)")

    return file


def find_xdg_cache_dir() -> Path:
    """Return the $XDG_CACHE_DIR.

    This is from the XDG Base Dir spec[1,2] which is standard on Linux.

    [1] - https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
    [2] - https://wiki.archlinux.org/title/XDG_Base_Directory
    """
    user_xdg_cache_home = os.environ.get("XDG_CACHE_HOME", None)

    default_xdg_cache_dir = Path.home() / ".cache"
    default_xdg_cache_origin = "$HOME/.cache"

    if user_xdg_cache_home is not None:
        cache_dir = Path(user_xdg_cache_home)
        origin = "$XDG_CACHE_HOME"
    else:
        cache_dir = default_xdg_cache_dir
        origin = default_xdg_cache_origin

    if not cache_dir.is_dir():
        try:
            cache_dir.mkdir(parents=True)
        except PermissionError:
            error.fatal(
                f"unable to create cache directory {origin} ({str(cache_dir)}): "
                f"permission denied"
            )

    return cache_dir



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from rattr import error

if TYPE_CHECKING:
    from rattr.config._types import Arguments


def _is_project_root(path: Path) -> bool:
    """Return `True` if the given file is a the project root."""
    if not path.is_dir():
        return False

    is_python_project = (path / "pyproject.toml").is_file()

    is_git_repo = (path / ".git").exists()
    is_mercurial_repo = (path / ".hg").is_dir()
    is_apache_svn_repo = (path / ".svn").is_dir()

    return is_python_project or is_git_repo or is_mercurial_repo or is_apache_svn_repo


def find_project_root() -> Path:
    """Return the project root, defaults to current working directory."""
    cwd = Path.cwd().resolve()

    if _is_project_root(cwd):
        return cwd

    for dir in (dir for dir in cwd.parents if _is_project_root(dir)):
        return dir

    return cwd


def find_pyproject_toml() -> Path | None:
    """Return the project's `pyproject.toml` file."""
    pyproject_toml = find_project_root() / "pyproject.toml"

    if pyproject_toml.is_file():
        return pyproject_toml

    return None


def validate_arguments(arguments: Arguments) -> Arguments:
    """Validate and return the given arguments."""
    if arguments._follow_imports_level == 0:
        error.rattr("follow imports not set, results likely to be incomplete")

    if arguments.is_strict and arguments.threshold != 0:
        error.rattr("rattr is in --strict mode, ignoring threshold")

    if arguments.threshold < 0:
        error.fatal("threshold must be a positive integer")

    if not arguments.target.is_file():
        error.fatal(f"file {str(arguments.target)!r} does not exist")

    if arguments.target.suffix != ".py":
        error.rattr(
            f"rattr target expects '*.py', got {str(arguments.target)!r}; "
            f"did you specify the right target?"
        )

    return arguments



"""Tests for module/file level features."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from unittest import mock

import pytest

from rattr.analyser.file import FileAnalyser
from rattr.models.context import compile_root_context
from rattr.models.symbol import CallInterface, Func, Name

if TYPE_CHECKING:
    from collections.abc import Iterator

    from tests.shared import StateFn


@pytest.fixture(autouse=True)
def __set_current_file(state: StateFn) -> Iterator[None]:
    with state(current_file=Path(__file__)):
        yield


class TestModuleLevelStatements:
    def test_assignment(self, parse):
        _ast = parse(
            """
            one = 1
            one = 1
            two = 2
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        assert results.ir_as_dict() == {}

    def test_typed_assignment(self, parse):
        _ast = parse(
            """
            one: int = 1
            one: int = 1
            two: int = 2
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        assert results.ir_as_dict() == {}

    def test_augmented_assignment(self, parse):
        _ast = parse(
            """
            one += 1
            one += 1
            two += 2
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        assert results.ir_as_dict() == {}

    def test_walrus_operator(self, parse):
        _ast = parse(
            """
            x = (y := 1)
            x = (y := 2)
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        assert results.ir_as_dict() == {}

    def test_multiple_assignment(self, parse):
        _ast = parse(
            """
            a, b = some_tuple
            x, y = 1, 2
            z = 3, 4
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        assert results.ir_as_dict() == {}

    def test_lambda(self, parse):
        # Anonymous
        _ast = parse(
            """
            lambda *a, **k: a.attr
            """
        )
        with mock.patch("sys.exit") as _exit:
            FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        # Named
        _ast = parse(
            """
            name = lambda *a, **k: a.attr
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        name = Func(name="name", interface=CallInterface(vararg="a", kwarg="k"))

        expected = {
            name: {
                "calls": set(),
                "dels": set(),
                "gets": {Name("a.attr", "a")},
                "sets": set(),
            },
        }

        assert results.ir_as_dict() == expected

    def test_walrus_multiple_assignment(self, parse):
        # Walrus multi assign w/ lambda
        _ast = parse(
            """
            other = (alpha, beta := 1)
            other, another = (alpha, beta := 1)
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        assert results.ir_as_dict() == {}

    def test_walrus_multiple_assignment_list(self, parse):
        # Walrus multi assign w/ lambda
        _ast = parse(
            """
            other = [alpha, beta := 1]
            other, another = [alpha, beta := 1]
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        assert results.ir_as_dict() == {}

    def test_walrus_lambda(self, parse):
        # Walrus'd Lambda
        _ast = parse(
            """
            other = (name := lambda *a, **k: a.attr)
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        name = Func(name="name", interface=CallInterface(vararg="a", kwarg="k"))
        other = Func(name="other", interface=CallInterface(vararg="a", kwarg="k"))

        expected = {
            name: {
                "calls": set(),
                "dels": set(),
                "gets": {Name("a.attr", "a")},
                "sets": set(),
            },
            other: {
                "calls": set(),
                "dels": set(),
                "gets": {Name("a.attr", "a")},
                "sets": set(),
            },
        }

        assert results.ir_as_dict() == expected

    def test_walrus_multiple_assignment_lambda(self, parse):
        # Walrus multi assign w/ lambda
        _ast = parse(
            """
            other = (alpha, beta := lambda *a, **k: a.attr)
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        beta = Func(name="beta", interface=CallInterface(vararg="a", kwarg="k"))

        expected = {
            beta: {
                "calls": set(),
                "dels": set(),
                "gets": {Name("a.attr", "a")},
                "sets": set(),
            },
        }

        assert results.ir_as_dict() == expected

    def test_walrus_multiple_assignment_lambda_list(self, parse):
        # Walrus multi assign w/ lambda
        _ast = parse(
            """
            other = [alpha, beta := lambda *a, **k: a.attr]
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        beta = Func(name="beta", interface=CallInterface(vararg="a", kwarg="k"))

        expected = {
            beta: {
                "calls": set(),
                "dels": set(),
                "gets": {Name("a.attr", "a")},
                "sets": set(),
            },
        }

        assert results.ir_as_dict() == expected



from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

import attrs

from rattr.models.ir import FileIr

if TYPE_CHECKING:
    from rattr.versioning.typing import TypeAlias


FileName: TypeAlias = str
ModuleName: TypeAlias = str

ImportIrs: TypeAlias = dict[ModuleName, FileIr]


class TargetIr(TypedDict):
    filename: FileName
    ir: FileIr


@attrs.frozen
class OutputIrs:
    import_irs: ImportIrs
    target_ir: TargetIr



from __future__ import annotations

from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING
from unittest import mock

import pytest

from rattr.models.ir import FileIr
from rattr.models.results.file import FileResults
from rattr.models.symbol import (
    Call,
    CallArguments,
    CallInterface,
    Class,
    Func,
    Name,
)
from rattr.models.util import serialise
from rattr.results import generate_results_from_ir
from tests.shared import Import_

if TYPE_CHECKING:
    from collections.abc import Iterator

    from tests.shared import FileIrFromDictFn, MakeRootContextFn, StateFn


@pytest.fixture(autouse=True)
def __set_current_file(state: StateFn) -> Iterator[None]:
    with state(current_file=Path(__file__)):
        yield


class TestResults:
    def test_generate_results_from_ir_no_calls(
        self,
        make_root_context: MakeRootContextFn,
    ):
        # No calls
        fn = Func(name="fn", interface=CallInterface(args=("arg",)))
        fn_ir = {
            "gets": {Name("arg.another_attr", "arg")},
            "sets": {Name("arg.attr", "arg")},
            "dels": set(),
            "calls": set(),
        }
        file_ir = FileIr(
            context=make_root_context([fn], include_root_symbols=True),
            file_ir={fn: fn_ir},
        )

        expected = FileResults(
            function_results={
                "fn": {
                    "gets": {"arg.another_attr"},
                    "sets": {"arg.attr"},
                    "dels": set(),
                    "calls": set(),
                }
            },
        )

        assert generate_results_from_ir(target_ir=file_ir, import_irs={}) == expected

    def test_generate_results_from_ir_simple(
        self,
        make_root_context: MakeRootContextFn,
    ):
        # Calls
        fn_a = Func(name="fn_a", interface=CallInterface(args=("arg",)))
        fn_b = Func(name="fn_b", interface=CallInterface(args=("arg_b",)))
        fn_a_ir = {
            "gets": {Name("arg.another_attr", "arg")},
            "sets": {Name("arg.attr", "arg")},
            "dels": set(),
            "calls": {
                Call(name="fn_b", args=CallArguments(args=("arg",)), target=fn_b),
            },
        }
        fn_b_ir = {
            "gets": {Name("arg_b.get_in_fn_b", "arg_b")},
            "sets": {Name("arg_b.set_in_fn_b", "arg_b")},
            "dels": set(),
            "calls": set(),
        }
        file_ir = FileIr(
            context=make_root_context([fn_a, fn_b], include_root_symbols=True),
            file_ir={
                fn_a: fn_a_ir,
                fn_b: fn_b_ir,
            },
        )

        expected = FileResults(
            function_results={
                "fn_a": {
                    "gets": {"arg.another_attr", "arg.get_in_fn_b"},
                    "sets": {"arg.attr", "arg.set_in_fn_b"},
                    "dels": set(),
                    "calls": {"fn_b()"},
                },
                "fn_b": {
                    "gets": {"arg_b.get_in_fn_b"},
                    "sets": {"arg_b.set_in_fn_b"},
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert generate_results_from_ir(target_ir=file_ir, import_irs={}) == expected

    def test_generate_results_from_ir_direct_recursion(
        self,
        make_root_context: MakeRootContextFn,
    ):
        # Direct recursion
        fn = Func(name="fn", interface=CallInterface(args=("arg",)))
        fn_ir = {
            "gets": {Name("arg.another_attr", "arg")},
            "sets": {Name("arg.attr", "arg")},
            "dels": set(),
            "calls": {Call(name="fn()", args=CallArguments(args=("arg",)))},
        }
        file_ir = FileIr(
            context=make_root_context([fn], include_root_symbols=True),
            file_ir={fn: fn_ir},
        )

        expected = FileResults(
            function_results={
                "fn": {
                    "gets": {"arg.another_attr"},
                    "sets": {"arg.attr"},
                    "dels": set(),
                    "calls": {"fn()"},
                }
            },
        )

        assert generate_results_from_ir(target_ir=file_ir, import_irs={}) == expected

    def test_generate_results_from_ir_indirect_recursion(
        self,
        make_root_context: MakeRootContextFn,
    ):
        # Indirect recursion
        fn_a = Func(name="fn_a", interface=CallInterface(args=("arg_a",)))
        fn_b = Func(name="fn_b", interface=CallInterface(args=("arg_b",)))
        fn_a_ir = {
            "gets": set(),
            "sets": {Name("arg_a.get_from_a", "arg_a")},
            "dels": set(),
            "calls": {
                Call(name="fn_b", args=CallArguments(args=("arg_a",)), target=fn_b),
            },
        }
        fn_b_ir = {
            "gets": set(),
            "sets": {Name("arg_b.get_from_b", "arg_b")},
            "dels": set(),
            "calls": {
                Call(name="fn_a", args=CallArguments(args=("arg_b",)), target=fn_a),
            },
        }
        file_ir = FileIr(
            context=make_root_context([fn_a, fn_b], include_root_symbols=True),
            file_ir={
                fn_a: fn_a_ir,
                fn_b: fn_b_ir,
            },
        )

        expected = FileResults(
            function_results={
                "fn_a": {
                    "gets": set(),
                    "sets": {"arg_a.get_from_a", "arg_a.get_from_b"},
                    "dels": set(),
                    "calls": {"fn_b()"},
                },
                "fn_b": {
                    "gets": set(),
                    "sets": {"arg_b.get_from_a", "arg_b.get_from_b"},
                    "dels": set(),
                    "calls": {"fn_a()"},
                },
            },
        )

        assert generate_results_from_ir(target_ir=file_ir, import_irs={}) == expected

    def test_generate_results_from_ir_child_has_direct_recursion(
        self,
        make_root_context: MakeRootContextFn,
    ):
        fn_a = Func(name="fn_a", interface=CallInterface(args=("x",)))
        fn_b = Func(name="fn_b", interface=CallInterface(args=("arg",)))
        fn_a_ir = {
            "gets": {Name("x.attr", "x")},
            "sets": set(),
            "dels": set(),
            "calls": {
                Call(name="fn_b", args=CallArguments(args=("x",)), target=fn_b),
            },
        }
        fn_b_ir = {
            "gets": {Name("arg.field", "arg")},
            "sets": set(),
            "dels": set(),
            "calls": {
                Call(name="fn_b", args=CallArguments(args=("arg",)), target=fn_b),
            },
        }
        file_ir = FileIr(
            context=make_root_context([fn_a, fn_b], include_root_symbols=True),
            file_ir={
                fn_a: fn_a_ir,
                fn_b: fn_b_ir,
            },
        )

        expected = FileResults(
            function_results={
                "fn_a": {
                    "gets": {"x.attr", "x.field"},
                    "sets": set(),
                    "dels": set(),
                    "calls": {"fn_b()"},
                },
                "fn_b": {
                    "gets": {"arg.field"},
                    "sets": set(),
                    "dels": set(),
                    "calls": {"fn_b()"},
                },
            },
        )

        assert generate_results_from_ir(target_ir=file_ir, import_irs={}) == expected

    def test_generate_results_from_ir_child_has_indirect_recursion(
        self,
        make_root_context: MakeRootContextFn,
    ):
        fn_a = Func(name="fn_a", interface=CallInterface(args=("a",)))
        fn_b = Func(name="fn_b", interface=CallInterface(args=("b",)))
        fn_c = Func(name="fn_c", interface=CallInterface(args=("c",)))
        fn_a_ir = {
            "gets": {Name("a.in_a", "a")},
            "sets": set(),
            "dels": set(),
            "calls": {
                Call(
                    name="fn_b",
                    args=CallArguments(args=("a",), kwargs={}),
                    target=fn_b,
                ),
            },
        }
        fn_b_ir = {
            "gets": {Name("b.in_b", "b")},
            "sets": set(),
            "dels": set(),
            "calls": {
                Call(
                    name="fn_c",
                    args=CallArguments(args=("b",), kwargs={}),
                    target=fn_c,
                ),
            },
        }
        fn_c_ir = {
            "gets": {Name("c.in_c", "c")},
            "sets": set(),
            "dels": set(),
            "calls": {
                Call(
                    name="fn_b",
                    args=CallArguments(args=(), kwargs={"b": "c"}),
                    target=fn_b,
                ),
            },
        }
        file_ir = FileIr(
            context=make_root_context([fn_a, fn_b, fn_c], include_root_symbols=True),
            file_ir={
                fn_a: fn_a_ir,
                fn_b: fn_b_ir,
                fn_c: fn_c_ir,
            },
        )

        expected = FileResults(
            function_results={
                "fn_a": {
                    "gets": {"a.in_a", "a.in_b", "a.in_c"},
                    "sets": set(),
                    "dels": set(),
                    "calls": {"fn_b()"},
                },
                "fn_b": {
                    "gets": {"b.in_b", "b.in_c"},
                    "sets": set(),
                    "dels": set(),
                    "calls": {"fn_c()"},
                },
                "fn_c": {
                    "gets": {"c.in_b", "c.in_c"},
                    "sets": set(),
                    "dels": set(),
                    "calls": {"fn_b()"},
                },
            },
        )

        assert generate_results_from_ir(target_ir=file_ir, import_irs={}) == expected

    def test_generate_results_from_ir_repeated_calls_should_be_ignored(
        self,
        make_root_context: MakeRootContextFn,
    ):
        fn_a = Func(name="fn_a", interface=CallInterface(args=("a",)))
        fn_b = Func(name="fn_b", interface=CallInterface(args=("b",)))
        fn_a_ir = {
            "gets": {Name("a.in_a", "a")},
            "sets": set(),
            "dels": set(),
            "calls": {
                Call(name="fn_b", args=CallArguments(args=("a.attr",)), target=fn_b),
                Call(name="fn_b", args=CallArguments(args=("a.attr",)), target=fn_b),
            },
        }
        fn_b_ir = {
            "gets": {Name("b.in_b", "b")},
            "sets": set(),
            "dels": set(),
            "calls": set(),
        }
        file_ir = FileIr(
            context=make_root_context([fn_a, fn_b], include_root_symbols=True),
            file_ir={
                fn_a: fn_a_ir,
                fn_b: fn_b_ir,
            },
        )

        expected = FileResults(
            function_results={
                "fn_a": {
                    "gets": {"a.in_a", "a.attr.in_b"},
                    "sets": set(),
                    "dels": set(),
                    "calls": {"fn_b()"},
                },
                "fn_b": {
                    "gets": {"b.in_b"},
                    "sets": set(),
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert generate_results_from_ir(target_ir=file_ir, import_irs={}) == expected

    def test_generate_results_from_ir_repeated_calls_should_not_be_ignored(
        self,
        make_root_context: MakeRootContextFn,
    ):
        # Should not be ignored as they have different call signatures
        fn_a = Func(name="fn_a", interface=CallInterface(args=("a",)))
        fn_b = Func(name="fn_b", interface=CallInterface(args=("b",)))
        fn_a_ir = {
            "gets": {Name("a.in_a", "a")},
            "sets": set(),
            "dels": set(),
            "calls": {
                Call(
                    name="fn_b",
                    args=CallArguments(args=("a.attr_one",)),
                    target=fn_b,
                ),
                Call(
                    name="fn_b",
                    args=CallArguments(args=("a.attr_two",)),
                    target=fn_b,
                ),
            },
        }
        fn_b_ir = {
            "gets": {Name("b.in_b", "b")},
            "sets": set(),
            "dels": set(),
            "calls": set(),
        }
        file_ir = FileIr(
            context=make_root_context([fn_a, fn_b], include_root_symbols=True),
            file_ir={
                fn_a: fn_a_ir,
                fn_b: fn_b_ir,
            },
        )

        expected = FileResults(
            function_results={
                "fn_a": {
                    "gets": {"a.in_a", "a.attr_one.in_b", "a.attr_two.in_b"},
                    "sets": set(),
                    "dels": set(),
                    "calls": {"fn_b()"},
                },
                "fn_b": {
                    "gets": {"b.in_b"},
                    "sets": set(),
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert generate_results_from_ir(target_ir=file_ir, import_irs={}) == expected

    def test_import_irs(
        self,
        file_ir_from_dict: FileIrFromDictFn,
        make_root_context: MakeRootContextFn,
    ):
        act = Func(name="act", interface=CallInterface(args=("arg",)))
        import_irs = {
            "module": file_ir_from_dict(
                {
                    act: {
                        "gets": {Name("arg.attr", "arg")},
                        "sets": set(),
                        "dels": set(),
                        "calls": set(),
                    }
                }
            )
        }

        import_ = Import_(
            name="act",
            qualified_name="module.act",
            module_name_and_spec=("module", mock.Mock()),
        )

        fn = Func(name="fn", interface=CallInterface(args=("ms",)))
        fn_ir = {
            "gets": set(),
            "sets": set(),
            "dels": set(),
            "calls": {
                Call(name="act", args=CallArguments(args=("ms",)), target=import_)
            },
        }
        file_ir = FileIr(
            context=make_root_context([import_, fn], include_root_symbols=True),
            file_ir={fn: fn_ir},
        )

        expected = FileResults(
            function_results={
                "fn": {
                    "gets": {"ms.attr"},
                    "sets": set(),
                    "dels": set(),
                    "calls": {"act()"},
                }
            },
        )

        assert (
            generate_results_from_ir(target_ir=file_ir, import_irs=import_irs)
            == expected
        )

    def test_import_irs_chained_import(
        self,
        file_ir_from_dict: FileIrFromDictFn,
        make_root_context: MakeRootContextFn,
    ):
        import_second = Import_(
            name="second",
            qualified_name="chained.second",
            module_name_and_spec=("chained", mock.Mock()),
        )

        first = Func(name="first", interface=CallInterface(args=("arrg",)))
        second = Func(name="second", interface=CallInterface(args=("blarg",)))
        import_irs = {
            "module": file_ir_from_dict(
                {
                    first: {
                        "gets": set(),
                        "sets": set(),
                        "dels": set(),
                        "calls": {
                            Call(
                                name="second",
                                args=CallArguments(args=("arrg",)),
                                target=import_second,
                            ),
                        },
                    }
                }
            ),
            "chained": file_ir_from_dict(
                {
                    second: {
                        "gets": {Name("blarg._attr", "blarg")},
                        "sets": set(),
                        "dels": set(),
                        "calls": set(),
                    }
                }
            ),
        }

        import_first = Import_(
            name="first",
            qualified_name="module.first",
            module_name_and_spec=("module", mock.Mock()),
        )

        fn = Func(name="fn", interface=CallInterface(args=("flarg",)))
        fn_ir = {
            "gets": set(),
            "sets": set(),
            "dels": set(),
            "calls": {
                Call(
                    name="first",
                    args=CallArguments(args=("flarg",)),
                    target=import_first,
                ),
            },
        }
        file_ir = FileIr(
            context=make_root_context([fn, import_first], include_root_symbols=True),
            file_ir={fn: fn_ir},
        )

        expected = FileResults(
            function_results={
                "fn": {
                    "gets": {"flarg._attr"},
                    "sets": set(),
                    "dels": set(),
                    "calls": {"first()"},
                }
            },
        )

        assert (
            generate_results_from_ir(target_ir=file_ir, import_irs=import_irs)
            == expected
        )

    def test_class(self, make_root_context: MakeRootContextFn):
        cls_inst = Class("SomeClass", interface=CallInterface(args=("self", "arg")))
        cls_inst_ir = {
            "gets": {Name("arg.attr_in_init", "arg")},
            "sets": {Name("self.my_attr", "self")},
            "dels": set(),
            "calls": set(),
        }
        cls_inst_sm = Func(
            name="SomeClass.static",
            interface=CallInterface(args=("flarg",)),
        )
        cls_inst_sm_ir = {
            "gets": set(),
            "sets": {Name("flarg.attr_in_static", "flarg")},
            "dels": set(),
            "calls": set(),
        }
        fn = Func(
            name="i_call_them",
            interface=CallInterface(args=("marg",)),
        )
        fn_ir = {
            "gets": set(),
            "sets": {Name("instance")},
            "dels": set(),
            "calls": {
                Call(
                    name="SomeClass()",
                    args=CallArguments(args=("instance", "marg")),
                    target=cls_inst,
                ),
                Call(
                    name="SomeClass.static()",
                    args=CallArguments(args=("marg",)),
                    target=cls_inst_sm,
                ),
            },
        }
        file_ir = FileIr(
            context=make_root_context(
                [
                    cls_inst,
                    cls_inst_sm,
                    fn,
                ],
                include_root_symbols=True,
            ),
            file_ir={
                cls_inst: cls_inst_ir,
                cls_inst_sm: cls_inst_sm_ir,
                fn: fn_ir,
            },
        )

        expected = FileResults(
            function_results={
                "SomeClass": {
                    "gets": {"arg.attr_in_init"},
                    "sets": {"self.my_attr"},
                    "dels": set(),
                    "calls": set(),
                },
                "SomeClass.static": {
                    "gets": set(),
                    "sets": {"flarg.attr_in_static"},
                    "dels": set(),
                    "calls": set(),
                },
                "i_call_them": {
                    "gets": {
                        "marg.attr_in_init",
                    },
                    "sets": {
                        "instance",
                        "instance.my_attr",
                        "marg.attr_in_static",
                    },
                    "dels": set(),
                    "calls": {
                        "SomeClass()",
                        "SomeClass.static()",
                    },
                },
            },
        )

        assert generate_results_from_ir(target_ir=file_ir, import_irs={}) == expected


class TestResultsEncoder:
    @pytest.fixture
    def example_file_results_unordered(self):
        return {
            "my_func": {
                "gets": {
                    "z_should_come_after_a",
                    "z_should_come_after_a.yet_it_did_not",
                    "a_should_come_before_z",
                    "a_should_come_before_z[]",
                    "argument.attrs[].whatever.item",
                    "@ListComp",
                    "@Str",
                },
                "sets": {
                    "z_should_come_after_a",
                    "a_should_come_before_z",
                },
                "dels": {
                    "a_should_come_before_z",
                    "z_should_come_after_a",
                },
                "calls": {
                    "my_amazing_function()",
                    "your_amazing_function()",
                    "library.your_amazing_function()",
                    "argument.callbacks[].exec()",
                },
            }
        }

    @pytest.fixture
    def example_file_results_ordered_str(self):
        return dedent(
            """\
            {
                "my_func": {
                    "gets": [
                        "@ListComp",
                        "@Str",
                        "a_should_come_before_z",
                        "a_should_come_before_z[]",
                        "argument.attrs[].whatever.item",
                        "z_should_come_after_a",
                        "z_should_come_after_a.yet_it_did_not"
                    ],
                    "sets": [
                        "a_should_come_before_z",
                        "z_should_come_after_a"
                    ],
                    "dels": [
                        "a_should_come_before_z",
                        "z_should_come_after_a"
                    ],
                    "calls": [
                        "argument.callbacks[].exec()",
                        "library.your_amazing_function()",
                        "my_amazing_function()",
                        "your_amazing_function()"
                    ]
                }
            }"""
        )

    def test_is_sorted_from_unordered_container(
        self,
        example_file_results_unordered,
        example_file_results_ordered_str,
    ):
        assert (
            serialise(FileResults(example_file_results_unordered), indent=4)
            == example_file_results_ordered_str
        )



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from rattr.ast.types import Identifier


def gen_import_from_stmt(module: Identifier, target: Identifier) -> str:
    if not module.isidentifier():
        raise ValueError(f"{module!r} is not a valid identifier")

    if target != "*" and not target.isidentifier():
        raise ValueError(f"{target!r} is not a valid identifier")

    return f"from {module} import {target}"



"""Annotations to be imported for use by Rattr analysed code."""
from __future__ import annotations

from typing import TYPE_CHECKING

from rattr.analyser.types import (
    KeywordArgumentName,
    LocalIdentifier,
    PositionalArgumentName,
    TargetName,
)
from rattr.ast.types import Identifier

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import TypeVar

    from rattr.versioning.typing import ParamSpec

    P = ParamSpec("P")
    R = TypeVar("R")


def rattr_ignore(*optional_target: Callable[P, R]) -> Callable[P, R]:
    """Mark the given function as ignored by rattr."""
    if optional_target:
        if len(optional_target) != 1:
            raise ValueError("rattr_ignore expects exactly one target")

        return optional_target[0]

    def _inner(target: Callable[P, R]) -> Callable[P, R]:
        return target

    return _inner


def rattr_results(
    *,
    gets: set[Identifier] | None = None,
    sets: set[Identifier] | None = None,
    dels: set[Identifier] | None = None,
    calls: list[
        tuple[
            TargetName,
            tuple[
                list[PositionalArgumentName],
                dict[KeywordArgumentName, LocalIdentifier],
            ],
        ]
    ]
    | None = None,
) -> Callable[P, R]:
    """Explicitly provide the expected rattr results for a function.

    Note that when results are explicitly given rattr will take these as true and will
    not analyse the function body in any way.

    Note also that `Identifier`, `TargetName`, `PositionalArgumentName`,
    `KeywordArgumentName`, and `LocalIdentifier` are all type aliases for `str`.
    """

    def _inner(f: Callable[P, R]) -> Callable[P, R]:
        return f

    return _inner



from __future__ import annotations

# isort: off
from rattr.results._types import (
    IrCall,
    IrCallTreeNode,
    IrEnvironment,
    IrTarget,
)

# isort: on
from rattr.results._find_call_target import find_call_target_and_ir
from rattr.results._simplify_utils import (
    construct_call_swaps,
    unbind_ir_with_call_swaps,
    unbind_name,
)
from rattr.results.util import (
    destructively_simplify_ir_call_tree,
    generate_results_from_ir,
    make_target_ir_call_tree,
)

__all__ = [
    "IrCall",
    "IrCallTreeNode",
    "IrEnvironment",
    "IrTarget",
    "find_call_target_and_ir",
    "construct_call_swaps",
    "unbind_ir_with_call_swaps",
    "unbind_name",
    "destructively_simplify_ir_call_tree",
    "generate_results_from_ir",
    "make_target_ir_call_tree",
]



from __future__ import annotations

import argparse
import sys as _sys


class ArgumentParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        exit_on_error = kwargs.pop("exit_on_error", True)
        # Make sure to remove 'exit_on_error' from 'kwargs'
        # when calling parent constructor since Python
        # 3.7 and 3.8 versions of 'ArgumentParser' don't
        # expect it.
        super(ArgumentParser, self).__init__(*args, **kwargs)
        # Save 'exit_on_error' as an instance attribute
        # AFTER calling parent constructor. This also
        # ensures that we overwrite default 'exit_on_error'
        # instance attributes for Python 3.9+ versions of
        # ArgumentParser.
        self.exit_on_error = exit_on_error

    def parse_known_args(self, args=None, namespace=None):
        if args is None:
            # args default to the system args
            args = _sys.argv[1:]
        else:
            # make sure that args are mutable
            args = list(args)

        # default Namespace built from parser defaults
        if namespace is None:
            namespace = argparse.Namespace()

        # add any action defaults that aren't present
        for action in self._actions:
            if action.dest is not argparse.SUPPRESS:
                if not hasattr(namespace, action.dest):
                    if action.default is not argparse.SUPPRESS:
                        setattr(namespace, action.dest, action.default)

        # add any parser defaults that aren't present
        for dest in self._defaults:
            if not hasattr(namespace, dest):
                setattr(namespace, dest, self._defaults[dest])

        # parse the arguments and exit if there are any errors
        if self.exit_on_error:
            try:
                namespace, args = self._parse_known_args(args, namespace)
            except argparse.ArgumentError as err:
                self.error(str(err))
        else:
            namespace, args = self._parse_known_args(args, namespace)

        if hasattr(namespace, argparse._UNRECOGNIZED_ARGS_ATTR):
            args.extend(getattr(namespace, argparse._UNRECOGNIZED_ARGS_ATTR))
            delattr(namespace, argparse._UNRECOGNIZED_ARGS_ATTR)
        return namespace, args

    def exit(self, status=0, message=None):
        if not self.exit_on_error:
            raise argparse.ArgumentError(None, message or "")

        if message:
            self._print_message(message, _sys.stderr)

        if self.exit_on_error:
            _sys.exit(status)



from __future__ import annotations

from rattr.models.symbol._symbol import (
    AnyCallInterface,
    CallArguments,
    CallInterface,
    Location,
    Symbol,
)
from rattr.models.symbol._symbols import (
    PYTHON_ATTR_ACCESS_BUILTINS,
    PYTHON_BUILTINS,
    PYTHON_LITERAL_BUILTINS,
    Builtin,
    Call,
    Class,
    Func,
    Import,
    Name,
)
from rattr.models.symbol._types import UserDefinedCallableSymbol
from rattr.models.symbol.util import without_call_brackets

__all__ = [
    "AnyCallInterface",
    "CallArguments",
    "CallInterface",
    "Location",
    "Symbol",
    "PYTHON_LITERAL_BUILTINS",
    "PYTHON_ATTR_ACCESS_BUILTINS",
    "PYTHON_BUILTINS",
    "Builtin",
    "Call",
    "Class",
    "Func",
    "Import",
    "Name",
    "UserDefinedCallableSymbol",
    "without_call_brackets",
]



from __future__ import annotations

import ast
import sys
from collections.abc import Iterable, Mapping
from contextlib import contextmanager
from importlib.util import find_spec
from os.path import dirname, join
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING
from unittest import mock

import pytest

from rattr.analyser.base import Assertor, CustomFunctionAnalyser
from rattr.analyser.file import FileAnalyser
from rattr.ast.types import Identifier
from rattr.config import Arguments, Config, Output, State
from rattr.models.context import Context, SymbolTable, compile_root_context
from rattr.models.ir import FileIr, FunctionIr
from rattr.models.results import FileResults
from rattr.models.symbol import (
    Builtin,
    Call,
    CallArguments,
    Name,
    Symbol,
    UserDefinedCallableSymbol,
)
from rattr.results import generate_results_from_ir

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator, Mapping
    from typing import Final, TypeVar

    from tests.shared import (
        ArgumentsFn,
        FileIrFromDictFn,
        MakeRootContextFn,
        MakeSymbolTableFn,
        OsDependentPathFn,
        ParseFn,
        ParseWithContextFn,
        SetTestingConfigFn,
        StateFn,
    )

    StrOrPath = TypeVar("StrOrPath", str, Path)


pytest_python_version_markers: Final = (
    (3, 9),
    (3, 10),
    (3, 11),
    (3, 12),
)


def __python_version_marker(major: int, minor: int) -> str:
    return f"python_{major}_{minor}"


def pytest_configure(config):
    config.addinivalue_line("addopts", "--strict-markers")

    config.addinivalue_line("markers", "pypy: mark test to run only under pypy")
    config.addinivalue_line("markers", "cpython: mark test to run only under cpython")

    for major, minor in pytest_python_version_markers:
        marker = __python_version_marker(major, minor)
        config.addinivalue_line(
            "markers",
            f"{marker}: mark test to run only under Python {major}.{minor}",
        )

    config.addinivalue_line("markers", "windows: mark test to run only under Windows")
    config.addinivalue_line("markers", "linux: mark test to run only under Linux")
    config.addinivalue_line("markers", "posix: mark test to run only under Posix")

    config.addinivalue_line(
        "markers",
        "update_expected_results: mark test that updates the expected test results for "
        "a benchmarking test, only run if the mark is explicitly given",
    )
    config.addinivalue_line(
        "markers",
        "update_expected_irs: as with update_expected_results but for irs",
    )


def pytest_collection_modifyitems(config: pytest.Config, items: list[pytest.Item]):
    """Alter the collected tests."""
    if not is_pypy():
        skip_test_items_with_mark(items, "pypy")
    if not is_cpython():
        skip_test_items_with_mark(items, "cpython")

    for major, minor in pytest_python_version_markers:
        if not is_python_version(f"{major}.{minor}"):
            skip_test_items_with_mark(items, __python_version_marker(major, minor))

    if not is_windows():
        skip_test_items_with_mark(items, "windows")
    if not is_linux():
        skip_test_items_with_mark(items, "linux")
    if not is_posix():
        skip_test_items_with_mark(items, "posix")

    skip_test_items_with_mark_if_not_explicitly_given(
        items,
        "update_expected_results",
        config=config,
    )
    skip_test_items_with_mark_if_not_explicitly_given(
        items,
        "update_expected_irs",
        config=config,
    )


def skip_test_items_with_mark(items: list[pytest.Item], mark: str):
    """Remove the marked items if the condition is `False`."""
    _reason = "{mark} is not satisfied"

    for item in (i for i in items if i.get_closest_marker(mark)):
        item.add_marker(pytest.mark.skip(reason=_reason.format(mark=mark)))


def skip_test_items_with_mark_if_not_explicitly_given(
    items: list[pytest.Item],
    mark: str,
    *,
    config: pytest.Config,
):
    """Skip tests with the given mark unless the mark was given via `-m mark` etc."""
    dash_m_is_in_cli_arguments = bool(config.option.markexpr)
    dash_k_is_in_cli_arguments = bool(config.option.keyword)

    if dash_m_is_in_cli_arguments or dash_k_is_in_cli_arguments:
        # If `-m ...` or `-k ...` is given, let pytest handle the skipping/running
        # i.e. the test will only run if it is in the `-m` or the `-k`
        return

    skip_test_items_with_mark(items, mark)


def is_pypy() -> bool:
    return find_spec("__pypy__") is not None


def is_cpython() -> bool:
    return not is_pypy()


def is_python_version(version: str) -> bool:
    return version == f"{sys.version_info.major}.{sys.version_info.minor}"


def is_windows() -> bool:
    return sys.platform == "win32"


def is_linux() -> bool:
    return sys.platform == "linux"


def is_posix() -> bool:
    if sys.platform == "linux":
        return True
    if sys.platform == "darwin":
        return True
    return False


@pytest.fixture(scope="session", autouse=True)
@mock.patch("rattr.config._types.validate_arguments", lambda args: args)
def _init_testing_config() -> None:
    Config(
        arguments=Arguments(
            pyproject_toml_override=None,
            _follow_imports_level=1,
            _excluded_imports=set(),
            _excluded_names=set(),
            _warning_level="default",
            collapse_home=True,
            truncate_deep_paths=True,
            is_strict=False,
            threshold=0,
            stdout=Output.results,
            target=Path("target.py"),
        ),
        state=State(),
    )


@pytest.fixture(scope="function")
@mock.patch("rattr.config._types.validate_arguments", lambda args: args)
def set_testing_config() -> SetTestingConfigFn:
    def _set_testing_config(arguments: Arguments, state: State = None) -> None:
        Config(arguments=arguments, state=state or State())

    return _set_testing_config


def _nth_line_is_empty(lines: list[str], *, n: int) -> bool:
    return lines and (lines[n] == "" or lines[n].isspace())


@pytest.fixture
def parse() -> ParseFn:
    def _inner(source: str) -> ast.Module:
        """Return the parsed AST for the given code, use relative indentation.

        Assumed usage is:
            parse(
                '''
                <source code>       # first line sets the base indent
                <source code>
                    <source code>   # this is indented once (relative)
                <source code>
                '''
            )

        """
        lines = source.splitlines()

        # Skip the first and last line if empty, this is because in the usage such as:
        # parse(
        #   """
        #   ...
        #   """
        # )
        # the first and last line (specifically the first) will have a different
        # indentation level, which will throw off `dedent(...)`.
        if _nth_line_is_empty(lines, n=0):
            lines = lines[1:]
        if _nth_line_is_empty(lines, n=-1):
            lines = lines[:-1]

        if not len(lines):
            raise ValueError("parse(...) expects a non-empty, non-whitespace string")

        return ast.parse(dedent("\n".join(lines)))

    return _inner


@pytest.fixture
def parse_with_context(parse: ParseFn):
    def _inner(source: str) -> ParseWithContextFn:
        ast_module = parse(source)
        context = compile_root_context(ast_module)
        return ast_module, context

    return _inner


@pytest.fixture
def analyse_single_file(
    parse_with_context: Callable[[str], tuple[ast.AST, Context]],
) -> Callable[[str], tuple[FileIr, FileResults]]:
    """Parse and analyse the source as though it were a single file.

    NOTE
        This does not follow imports, for that we will need a more complex fixture or to
        wait for the new config code to be merged s.t. we can mock the config easier and
        call the functions used in __main__.py more directly.
    """

    def _inner(source: str) -> tuple[ast.AST, Context]:
        ast_module, context = parse_with_context(source)
        file_ir = FileAnalyser(ast_module, context).analyse()
        file_results = generate_results_from_ir(target_ir=file_ir, import_irs={})

        return file_ir, file_results

    return _inner


@pytest.fixture
def builtin() -> Callable[[str], Builtin]:
    # TODO This is no longer useful, refactor to remove it

    def _inner(name: str) -> Builtin:
        return Builtin(name)

    return _inner


@pytest.fixture()
def run_in_strict_mode(arguments: ArgumentsFn) -> Iterator[None]:
    with arguments(is_strict=True):
        yield


@pytest.fixture()
def run_in_permissive_mode(arguments: ArgumentsFn) -> Iterator[None]:
    with arguments(is_strict=False):
        yield


@pytest.fixture
def make_symbol_table(make_root_context: MakeRootContextFn) -> MakeSymbolTableFn:
    def _make_symbol_table(
        symbols: Mapping[Identifier, Symbol] | Iterable[Symbol] = (),
        *,
        include_root_symbols: bool = False,
    ) -> SymbolTable:
        context = make_root_context(symbols, include_root_symbols=include_root_symbols)
        return context.symbol_table

    return _make_symbol_table


@pytest.fixture
def make_root_context() -> MakeRootContextFn:
    def _make_root_context(
        symbols: Mapping[Identifier, Symbol] | Iterable[Symbol] = (),
        *,
        include_root_symbols: bool = False,
    ) -> SymbolTable:
        if include_root_symbols:
            context = compile_root_context(ast.Module(body=[]))
        else:
            context = Context(parent=None)

        if isinstance(symbols, Mapping):
            context.symbol_table._symbols = symbols
        elif isinstance(symbols, Iterable):
            context.add(symbols)
        else:
            raise TypeError

        return context

    return _make_root_context


@pytest.fixture
def arguments() -> ArgumentsFn:
    @contextmanager
    def _inner(**kwargs):
        arguments = Config().arguments

        _missing_attrs = {
            attr for attr in kwargs.keys() if not hasattr(arguments, attr)
        }
        if _missing_attrs:
            raise AttributeError(_missing_attrs)

        previous = {attr: getattr(arguments, attr) for attr in kwargs.keys()}

        for attr, value in kwargs.items():
            setattr(arguments, attr, value)

        yield

        for attr, value in previous.items():
            setattr(arguments, attr, value)

    return _inner


@pytest.fixture
def state() -> StateFn:
    @contextmanager
    def _inner(**kwargs):
        state = Config().state

        _missing_attrs = {attr for attr in kwargs.keys() if not hasattr(state, attr)}
        if _missing_attrs:
            raise AttributeError(_missing_attrs)

        previous = {attr: getattr(state, attr) for attr in kwargs.keys()}

        for attr, value in kwargs.items():
            setattr(state, attr, value)

        yield

        for attr, value in previous.items():
            setattr(state, attr, value)

    return _inner


@pytest.fixture
def config() -> Config:
    return Config()


@pytest.fixture
def stdlib_modules() -> set[str]:
    # Scraped from python.org
    scraped = {
        "string",
        "re",
        "difflib",
        "textwrap",
        "unicodedata",
        "stringprep",
        "rlcompleter",
        "struct",
        "codecs",
        "datetime",
        "calendar",
        "collections",
        "abc",
        "heapq",
        "bisect",
        "array",
        "weakref",
        "types",
        "copy",
        "pprint",
        "reprlib",
        "enum",
        "numbers",
        "math",
        "cmath",
        "decimal",
        "fractions",
        "random",
        "statistics",
        "itertools",
        "functools",
        "operator",
        "pathlib",
        "fileinput",
        "stat",
        "filecmp",
        "tempfile",
        "glob",
        "fnmatch",
        "linecache",
        "shutil",
        "pickle",
        "copyreg",
        "shelve",
        "dbm",
        "sqlite3",
        "zlib",
        "gzip",
        "bz2",
        "lzma",
        "zipfile",
        "tarfile",
        "csv",
        "configparser",
        "netrc",
        "xdrlib",
        "plistlib",
        "hashlib",
        "hmac",
        "secrets",
        "os",
        "io",
        "time",
        "argparse",
        "getopt",
        "logging",
        "getpass",
        "curses",
        "platform",
        "errno",
        "ctypes",
        "threading",
        "multiprocessing",
        "concurrent",
        "subprocess",
        "sched",
        "queue",
        "_thread",
        "asyncio",
        "socket",
        "ssl",
        "select",
        "selectors",
        "asyncore",
        "asynchat",
        "signal",
        "mmap",
        "email",
        "json",
        "mailcap",
        "mailbox",
        "mimetypes",
        "base64",
        "binhex",
        "binascii",
        "quopri",
        "uu",
        "html",
        "webbrowser",
        "cgi",
        "cgitb",
        "wsgiref",
        "urllib",
        "http",
        "ftplib",
        "poplib",
        "imaplib",
        "nntplib",
        "smtplib",
        "smtpd",
        "telnetlib",
        "uuid",
        "socketserver",
        "xmlrpc",
        "ipaddress",
        "audioop",
        "aifc",
        "sunau",
        "wave",
        "chunk",
        "colorsys",
        "imghdr",
        "sndhdr",
        "ossaudiodev",
        "gettext",
        "locale",
        "turtle",
        "cmd",
        "shlex",
        "typing",
        "pydoc",
        "doctest",
        "unittest",
        "test",
        "bdb",
        "faulthandler",
        "pdb",
        "timeit",
        "trace",
        "tracemalloc",
        "venv",
        "zipapp",
        "sys",
        "sysconfig",
        "builtins",
        "warnings",
        "dataclasses",
        "contextlib",
        "atexit",
        "traceback",
        "gc",
        "inspect",
        "site",
        "code",
        "codeop",
        "pkgutil",
        "modulefinder",
        "runpy",
        "importlib",
        "ast",
        "symtable",
        "token",
        "keyword",
        "tokenize",
        "tabnanny",
        "pyclbr",
        "py_compile",
        "compileall",
        "dis",
        "pickletools",
        "spwd",
        "crypt",
        "tty",
        "pty",
        "pipes",
        "nis",
        "optparse",
        "imp",
    }

    return scraped


@pytest.fixture
def builtins() -> set[str]:
    generated = {
        "abs",
        "all",
        "any",
        "ascii",
        "bin",
        "bool",
        "breakpoint",
        "bytearray",
        "bytes",
        "callable",
        "chr",
        "classmethod",
        "compile",
        "complex",
        "copyright",
        "credits",
        "delattr",
        "dict",
        "dir",
        "divmod",
        "enumerate",
        "eval",
        "exec",
        "filter",
        "float",
        "format",
        "frozenset",
        "getattr",
        "globals",
        "hasattr",
        "hash",
        "help",
        "hex",
        "id",
        "input",
        "int",
        "isinstance",
        "issubclass",
        "iter",
        "len",
        "license",
        "list",
        "locals",
        "map",
        "max",
        "memoryview",
        "min",
        "next",
        "object",
        "oct",
        "open",
        "ord",
        "pow",
        "print",
        "property",
        "range",
        "repr",
        "reversed",
        "round",
        "set",
        "setattr",
        "slice",
        "sorted",
        "staticmethod",
        "str",
        "sum",
        "super",
        "tuple",
        "type",
        "vars",
        "zip",
    }

    return generated


@pytest.fixture
def snippet():
    def _inner(relative_path: str):
        return join(dirname(__file__), "snippets", relative_path)

    return _inner


@pytest.fixture
def file_ir_from_dict(make_root_context: MakeRootContextFn) -> FileIrFromDictFn:
    def _inner(ir: Mapping[UserDefinedCallableSymbol, FunctionIr]) -> FileIr:
        return FileIr(
            context=make_root_context(ir.keys(), include_root_symbols=False),
            file_ir=ir,
        )

    return _inner


class _PrintBuiltinAnalyser(CustomFunctionAnalyser):
    @property
    def name(self) -> str:
        return "print"

    @property
    def qualified_name(self) -> str:
        return "print"

    def on_def(
        self,
        name: str,
        node: ast.FunctionDef | ast.AsyncFunctionDef,
        ctx: Context,
    ) -> FunctionIr:
        return {
            "sets": {Name(name="set_in_print_def")},
            "gets": {Name(name="get_in_print_def")},
            "dels": {Name(name="del_in_print_def")},
            "calls": {
                Call(
                    name="call_in_print_def",
                    args=CallArguments(args=(), kwargs={}),
                    target=None,
                ),
            },
        }

    def on_call(self, name: str, node: ast.Call, ctx: Context) -> FunctionIr:
        return {
            "sets": {Name(name="set_in_print")},
            "gets": {Name(name="get_in_print")},
            "dels": {Name(name="del_in_print")},
            "calls": {
                Call(
                    name="call_in_print",
                    args=CallArguments(args=(), kwargs={}),
                    target=None,
                ),
            },
        }


@pytest.fixture
def builtin_print_analyser() -> _PrintBuiltinAnalyser:
    return _PrintBuiltinAnalyser()


class _ExampleFuncAnalyser(CustomFunctionAnalyser):
    @property
    def name(self) -> str:
        return "example"

    @property
    def qualified_name(self) -> str:
        return "module.example"

    def on_def(
        self,
        name: str,
        node: ast.FunctionDef | ast.AsyncFunctionDef,
        ctx: Context,
    ) -> FunctionIr:
        return {
            "sets": {Name(name="set_in_example_def")},
            "gets": {Name(name="get_in_example_def")},
            "dels": {Name(name="del_in_example_def")},
            "calls": {
                Call(
                    name="call_in_example_def",
                    args=CallArguments(args=(), kwargs={}),
                    target=None,
                ),
            },
        }

    def on_call(self, name: str, node: ast.Call, ctx: Context) -> FunctionIr:
        return {
            "sets": {Name(name="set_in_example")},
            "gets": {Name(name="get_in_example")},
            "dels": {Name(name="del_in_example")},
            "calls": {
                Call(
                    name="call_in_example",
                    args=CallArguments(args=(), kwargs={}),
                    target=None,
                ),
            },
        }


@pytest.fixture
def example_func_analyser() -> _ExampleFuncAnalyser:
    return _ExampleFuncAnalyser()


class _ExampleAssertor(Assertor):
    ...


@pytest.fixture
def example_assertor() -> _ExampleAssertor:
    return _ExampleAssertor()


@pytest.fixture
def constant() -> str:
    config = Config()

    _prefix = config.LITERAL_VALUE_PREFIX
    _constant = ast.Constant.__name__

    return f"{_prefix}{_constant}"


@pytest.fixture
def literal():
    config = Config()

    _prefix = config.LITERAL_VALUE_PREFIX

    def _inner(node: ast.AST | type[ast.AST]) -> str:
        if isinstance(node, ast.AST):
            cls = node.__class__
        else:
            cls = node

        return f"{_prefix}{cls.__name__}"

    return _inner


@pytest.fixture
def walrus():
    def parse(expr: str):
        return ast.parse(f"a = ({expr})").body[0].value

    def _inner(*exprs):
        if len(exprs) == 0:
            raise ValueError

        if len(exprs) == 1:
            return parse(exprs[0])
        else:
            return [parse(e) for e in exprs]

    return _inner


@pytest.fixture
def stringify_nodes():
    def _inner(nodes: ast.AST):
        return [ast.dump(n) for n in nodes]

    return _inner


@pytest.fixture
def os_dependent_path() -> OsDependentPathFn:
    def _make_path_os_independent(posix_style_path: StrOrPath) -> StrOrPath:
        if sys.platform != "win32":
            if isinstance(posix_style_path, str):
                return posix_style_path.replace("\\", "/")
            else:
                return Path(str(posix_style_path).replace("\\", "/"))
        else:
            if isinstance(posix_style_path, str):
                return posix_style_path.replace("/", "\\")
            else:
                return Path(str(posix_style_path).replace("/", "\\"))

    return _make_path_os_independent



#!/usr/bin/env python3
"""Rattr entry point."""
from __future__ import annotations

from math import log10
from typing import TYPE_CHECKING

from rattr import error
from rattr.analyser.file import RattrStats, parse_and_analyse_file
from rattr.analyser.types import ImportIrs
from rattr.cli import parse_arguments
from rattr.config import Config, Output, State
from rattr.models.ir import FileIr
from rattr.models.results import FileResults, make_cacheable_results
from rattr.models.util import serialise, serialise_irs
from rattr.results import generate_results_from_ir

if TYPE_CHECKING:
    from pathlib import Path

    from rattr.models.results import CacheableResults


def _init_rattr_config() -> Config:
    return Config(arguments=parse_arguments(), state=State())


def main(config: Config) -> None:
    """Rattr entry point."""
    file_ir, import_irs, stats = parse_and_analyse_file()
    results = generate_results_from_ir(target_ir=file_ir, import_irs=import_irs)

    if not config.is_within_badness_threshold:
        badness, threshold = config.state.badness, config.arguments.threshold
        error.fatal(f"exceeded allowed badness ({badness} > {threshold})")

    if config.arguments.stdout == Output.ir:
        show_ir(config.arguments.target, file_ir, import_irs)

    if config.arguments.stdout == Output.results:
        show_results(results)

    if config.arguments.stdout == Output.cacheable:
        show_cacheable_results(make_cacheable_results(results, file_ir, import_irs))

    if config.arguments.stdout == Output.stats:
        show_stats(stats)


def show_ir(file: Path, file_ir: FileIr, import_irs: ImportIrs) -> None:
    """Prettily print the given file and imports IR."""
    serialised = serialise_irs(
        target_name=str(file),
        target_ir=file_ir,
        import_irs=import_irs,
    )
    print(serialised)


def show_cacheable_results(results: CacheableResults) -> None:
    """Prettily print the given file results."""
    print(serialise(results, indent=4))


def show_results(results: FileResults) -> None:
    """Prettily print the given file results."""
    print(serialise(results, indent=4))


def show_stats(stats: RattrStats) -> None:
    """Prettily print the collected stats in tables."""
    row = "{:26} | {:18}"
    config = Config()

    # Collate time stats
    table_header = row.format("", "Time (Seconds)")
    table_width = len(table_header)
    times = {
        "Parse <file>": stats.parse_time,
        "Build root context": stats.root_context_time,
        "Run assertor": stats.assert_time,
        "Parse / analyse imports": stats.analyse_imports_time,
        "Analyse <file>": stats.analyse_file_time,
    }

    # Print time stats
    print(end="\n\n")
    print(table_header)
    print("=" * table_width)
    for col_one, col_two in times.items():
        print(row.format(col_one, format(col_two, ".9f") + " s"))

    # Collate imports stats
    table_header = row.format("", "# of Imports")
    table_width = len(table_header)
    imports = {
        "Total number of imports": stats.number_of_imports,
        "... of which unique": stats.number_of_unique_imports,
    }

    # Print imports stats
    print(end="\n\n")
    print(table_header)
    print("=" * table_width)
    digits = 1 + int(log10(max(*imports.values(), 1)))
    for col_one, col_two in imports.items():
        print(row.format(col_one, format(col_two).zfill(digits)))

    # Collate line stats
    table_header = row.format("", "# of Lines")
    table_width = len(table_header)
    lines = {
        "<file>": stats.file_lines,
        "Imports": stats.import_lines,
    }

    # Print line stats
    print(end="\n\n")
    print(table_header)
    print("=" * table_width)
    digits = 1 + int(log10(max(*lines.values())))
    for col_one, col_two in lines.items():
        print(row.format(col_one, format(col_two).zfill(digits)))

    # Collate badness stats
    table_header = row.format("", "Magnitude")
    table_width = len(table_header)
    badness_stats = {
        "Total badness": config.state.full_badness,
        "... from <file>": config.state.badness_from_target_file,
        "... from imports": config.state.badness_from_imports,
        "... from simplification": config.state.badness_from_simplification,
    }

    # Print badness stats
    print(end="\n\n")
    print(table_header)
    print("=" * table_width)
    if max(*badness_stats.values()) > 0:
        digits = 1 + int(log10(max(*badness_stats.values())))
    else:
        digits = 1
    for col_one, col_two in badness_stats.items():
        print(row.format(col_one, format(col_two).zfill(digits)))

    # Badness summary
    threshold_or_inf = (
        str(config.arguments.threshold) if config.arguments.threshold > 0 else "∞"
    )
    avg_badness_per_line = config.state.badness / stats.file_lines

    badness_summary_stats = {
        "True badness": config.state.badness,
        "Threshold": threshold_or_inf,
        "Average badness/line": format(avg_badness_per_line, ".3f")
        + " b/l (true badness / <file> lines)",
    }
    summary = f"{{:{1 + max(map(len, badness_summary_stats.keys()))}}}: {{}}"
    print(end="\n\n")
    for desc, stat in badness_summary_stats.items():
        print(summary.format(desc, stat))

    # Summary stats
    summary_stats = {
        "Total time": format(sum(times.values()), ".6f") + " s",
        "Total lines": sum(lines.values()),
        "Average lines/second": format(sum(lines.values()) / sum(times.values()), ".0f")
        + " l/s",
    }
    summary = f"{{:{1 + max(map(len, summary_stats.keys()))}}}: {{}}"
    print(end="\n\n")
    for desc, stat in summary_stats.items():
        print(summary.format(desc, stat))

    print(end="\n\n")


def entry_point():
    """Entry point for command line app."""
    main(_init_rattr_config())


if __name__ == "__main__":
    entry_point()



from __future__ import annotations

from typing import TYPE_CHECKING

from rattr import error

if TYPE_CHECKING:
    from typing import NoReturn

    from rattr.config._types import Arguments


def validate_arguments(arguments: Arguments) -> Arguments | NoReturn:
    if arguments._follow_imports_level == 0:
        error.rattr("follow imports not set, results likely to be incomplete")

    if arguments.is_strict and arguments.threshold != 0:
        error.rattr("rattr is in --strict mode, ignoring threshold")

    if arguments.threshold < 0:
        error.fatal("threshold must be a positive integer")

    if not arguments.target.is_file():
        error.fatal(f"file {str(arguments.target)!r} does not exist")

    if arguments.target.suffix != ".py":
        error.rattr(
            f"rattr target expects '*.py', got {str(arguments.target)!r}; "
            f"did you specify the right target?"
        )

    return arguments



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from collections.abc import Iterator

    from tests.shared import StateFn


@pytest.fixture()
def test_file() -> Path:
    return Path("test.py")


@pytest.fixture(autouse=True)
def __set_current_file(state: StateFn, test_file: Path) -> Iterator[None]:
    # Many symbols automatically derive a location
    # The derivation will error if the state's file is not set
    # Se just set it universally for all symbol tests.
    with state(current_file=test_file):
        yield



from __future__ import annotations

import argparse
from pathlib import Path
from unittest import mock

import pytest

from rattr.cli.parser import _parse_project_config, parse_arguments
from rattr.config import Arguments, Output


class TestToml:
    def test_illegal_field_is_filtered(
        self, required_sys_args, toml_with_illegal_field, illegal_field_name
    ):
        arguments = parse_arguments(
            sys_args=required_sys_args,
            project_toml_conf=toml_with_illegal_field,
            exit_on_error=False,
        )
        assert hasattr(arguments, illegal_field_name) is False

    def test_valid_toml(self, required_sys_args, toml_well_formed):
        arguments = parse_arguments(
            sys_args=required_sys_args,
            project_toml_conf=toml_well_formed,
            exit_on_error=False,
        )
        assert arguments == Arguments(
            # From toml
            _follow_imports_level=3,
            _excluded_imports=["a\\.b\\.c", "a\\.b.*", "a\\.b\\.c\\.e", "a\\.b\\.c.*"],
            _excluded_names=["a_.*", "b_.*", "_.*"],
            _warning_level="all",
            threshold=1,
            # From sys_args
            target=Path("my/rattr/target.py"),
            # Defaults
            pyproject_toml_override=None,
            collapse_home=False,
            truncate_deep_paths=False,
            is_strict=False,
            stdout=Output.results,
        )

    def test_valid_toml_without_sys_args(self, toml_well_formed):
        with pytest.raises(argparse.ArgumentError) as argument_error:
            parse_arguments(
                sys_args=[],
                project_toml_conf=toml_well_formed,
                exit_on_error=False,
            )

        assert (
            argument_error.value.message
            == "rattr: error: the following arguments are required: <file>\n"
        )

    def test_toml_is_overwritten_by_sys_args(self, required_sys_args):
        sys_args = [
            "--follow-imports",
            "3",
            "--threshold",
            "8",
            "--exclude",
            "fn_excluded_1",
            "--exclude",
            "fn_excluded_2",
            "--exclude",
            "fn_excluded_3",
            "this/is/the/target.py",
        ]
        toml = {"threshold": 500, "exclude": ["fn_excluded_4", "fn_excluded_5"]}

        # Sys only
        arguments = parse_arguments(
            sys_args=sys_args,
            project_toml_conf={},
            exit_on_error=False,
        )
        assert arguments == Arguments(
            # Defaults
            pyproject_toml_override=None,
            _warning_level="default",
            _excluded_imports=None,
            is_strict=False,
            threshold=8,
            collapse_home=False,
            truncate_deep_paths=False,
            stdout=Output.results,
            # Sys args
            _follow_imports_level=3,
            _excluded_names=["fn_excluded_1", "fn_excluded_2", "fn_excluded_3"],
            target=Path("this/is/the/target.py"),
        )

        # Toml only
        arguments = parse_arguments(
            sys_args=required_sys_args,
            project_toml_conf=toml,
            exit_on_error=False,
        )
        assert arguments == Arguments(
            # Defaults
            pyproject_toml_override=None,
            _follow_imports_level=1,
            _warning_level="default",
            _excluded_imports=None,
            collapse_home=False,
            truncate_deep_paths=False,
            is_strict=False,
            stdout=Output.results,
            # Toml
            _excluded_names=["fn_excluded_4", "fn_excluded_5"],
            threshold=500,
            # Required sys arg
            target=Path("my/rattr/target.py"),
        )

        # Both
        arguments = parse_arguments(
            sys_args=sys_args,
            project_toml_conf=toml,
            exit_on_error=False,
        )
        assert arguments == Arguments(
            # Defaults
            pyproject_toml_override=None,
            _follow_imports_level=3,
            _warning_level="default",
            _excluded_imports=None,
            collapse_home=False,
            truncate_deep_paths=False,
            is_strict=False,
            stdout=Output.results,
            # From toml and sys args
            _excluded_names=[
                "fn_excluded_4",
                "fn_excluded_5",
                "fn_excluded_1",
                "fn_excluded_2",
                "fn_excluded_3",
            ],
            # From sys args
            threshold=8,  # 500 from toml, overwritten by sys args
            target=Path("this/is/the/target.py"),
        )


class TestTomlValidation:
    @pytest.mark.parametrize(
        ("toml,expected_type,actual_type"),
        [
            ({"follow-imports": "1"}, "int", "str"),
            ({"follow-imports": ["1"]}, "int", "list[str]"),
            ({"exclude-imports": "1"}, "list[str]", "str"),
            ({"exclude-imports": [1, 2]}, "list[str]", "list[int]"),
            ({"exclude": "1"}, "list[str]", "str"),
            ({"exclude": [1, "2"]}, "list[str]", "list[int | str]"),
            ({"warning-level": 1}, "str", "int"),
            ({"warning-level": True}, "str", "bool"),
            ({"collapse-home": "1"}, "bool", "str"),
            ({"collapse-home": 1}, "bool", "int"),
            ({"truncate-deep-paths": ["1", "two"]}, "bool", "list[str]"),
            ({"truncate-deep-paths": [1, 2, None]}, "bool", "list[NoneType | int]"),
            ({"strict": "True"}, "bool", "str"),
            ({"strict": None}, "bool", "NoneType"),
            ({"threshold": 1.0}, "int", "float"),
            ({"threshold": None}, "int", "NoneType"),
            ({"stdout": 1.0}, "str", "float"),
            ({"stdout": None}, "str", "NoneType"),
        ],
    )
    def test_toml_type_error(self, required_sys_args, toml, expected_type, actual_type):
        for _field, _value in toml.items():
            field, value = _field, _value

        with pytest.raises(argparse.ArgumentError) as argument_error:
            parse_arguments(
                sys_args=required_sys_args,
                project_toml_conf=toml,
                exit_on_error=False,
            )

        _type_error = (
            "{arg!r} expects type {expected_type}, got {value!r} of type {actual_type}"
        )
        assert argument_error.value.message == _type_error.format(
            arg=field,
            expected_type=expected_type,
            value=value,
            actual_type=actual_type,
        )

    @pytest.mark.parametrize(
        ("toml,error"),
        [
            ({"follow-imports": 100}, "invalid choice: 100 (choose from 0, 1, 2, 3)"),
            (
                {"warning-level": "this-is-nonsense-but-well-typed"},
                "invalid choice: 'this-is-nonsense-but-well-typed' (choose from 'none', 'local', 'default', 'all')",
            ),
            (
                {"stdout": "this-is-nonsense-but-well-typed"},
                "invalid Output value: 'this-is-nonsense-but-well-typed'",
            ),
        ],
    )
    def test_toml_type_is_correct_and_value_is_invalid(
        self, required_sys_args, toml, error
    ):
        with pytest.raises(argparse.ArgumentError) as argument_error:
            parse_arguments(
                sys_args=required_sys_args,
                project_toml_conf=toml,
                exit_on_error=False,
            )

        assert argument_error.value.message == error

    @pytest.mark.parametrize(
        ("toml"),
        [
            ({"follow-imports": 0}),
            ({"follow-imports": 3}),
            ({"exclude-imports": ["time"]}),
            ({"exclude-imports": ["time", "os"]}),
            ({"exclude": ["my_func"]}),
            ({"exclude": ["your_func", "our_func"]}),
            ({"warning-level": "all"}),
            ({"warning-level": "default"}),
            ({"collapse-home": True}),
            ({"collapse-home": False}),
            ({"truncate-deep-paths": True}),
            ({"truncate-deep-paths": False}),
            ({"strict": True}),
            ({"strict": False}),
            ({"threshold": 1}),
            ({"threshold": -600}),  # Invalid but checked by `validate_arguments`
            ({"stdout": "ir"}),
            ({"stdout": "results"}),
        ],
    )
    def test_toml_type_is_correct_and_value_is_valid(self, required_sys_args, toml):
        arguments = parse_arguments(
            sys_args=required_sys_args,
            project_toml_conf=toml,
            exit_on_error=False,
        )
        assert str(arguments.target.as_posix()) == "my/rattr/target.py"


class TestTomlOverride:
    @mock.patch("rattr.cli.parser._toml_error")
    @mock.patch("rattr.cli.parser.parse_project_toml")
    def test_parse_project_config_from_input_config(
        self,
        m_parse_project_toml,
        m_toml_error,
        illegal_field_name,
        toml_override_path,
        toml_override,
    ):
        # Test 1 -- input config given, should just be validated and returned
        input_config = {
            "threshold": 1,
            "strict": True,
            illegal_field_name: "skip-me",
        }

        input_config_valid_only = input_config.copy()
        _ = input_config_valid_only.pop(illegal_field_name)

        assert (
            _parse_project_config(input_config=input_config, project_toml_override=None)
            == input_config_valid_only
        )
        assert (
            _parse_project_config(
                input_config=input_config, project_toml_override=toml_override_path
            )
            == input_config_valid_only
            != toml_override
        )

        assert not m_toml_error.called
        assert not m_parse_project_toml.called

    @mock.patch("rattr.cli.parser._toml_error")
    @mock.patch("rattr.cli.parser.find_pyproject_toml")
    def test_parse_project_config_from_pyproject_toml(
        self,
        m_find_pyproject_toml,
        m_toml_error,
        toml_well_formed_path,
        toml_well_formed,
    ):
        m_pyproject_toml = mock.Mock(spec=Path)
        m_pyproject_toml.is_file.return_value = True
        m_pyproject_toml.read_text.return_value = toml_well_formed_path.read_text()

        m_find_pyproject_toml.return_value = m_pyproject_toml

        assert (
            _parse_project_config(input_config={}, project_toml_override=None)
            == toml_well_formed
        )
        assert not m_toml_error.called

    @mock.patch("rattr.cli.parser._toml_error")
    @mock.patch("rattr.cli.parser.find_pyproject_toml")
    def test_parse_project_config_from_override(
        self,
        m_find_pyproject_toml,
        m_toml_error,
        toml_well_formed_path,
        toml_well_formed,
        toml_override_path,
        toml_override,
    ):
        m_pyproject_toml = mock.Mock(spec=Path)
        m_pyproject_toml.is_file.return_value = True
        m_pyproject_toml.read_text.return_value = toml_well_formed_path.read_text()

        m_find_pyproject_toml.return_value = m_pyproject_toml

        assert (
            _parse_project_config(
                input_config={}, project_toml_override=toml_override_path
            )
            == toml_override
            != toml_well_formed
        )
        assert not m_toml_error.called



from __future__ import annotations

from typing import TYPE_CHECKING

import attrs
from attrs import field

from rattr.analyser.base import Assertor, CustomFunctionAnalyser
from rattr.config import Config
from rattr.models.symbol import Builtin, Import, Symbol
from rattr.plugins.analysers import DEFAULT_FUNCTION_ANALYSERS

if TYPE_CHECKING:
    from collections.abc import Iterable

    from rattr.ast.types import FullyQualifiedName, Identifier, ModuleName


@attrs.mutable
class Plugins:
    _assertors: list[Assertor] = field(alias="assertors", converter=list)
    _analysers: list[CustomFunctionAnalyser] = field(alias="analysers", converter=list)

    _analysers_by_fully_qualified_name: dict[
        FullyQualifiedName,
        CustomFunctionAnalyser,
    ] = field(init=False)

    @property
    def assertors(self) -> list[Assertor]:
        return self._assertors

    @property
    def analysers(self) -> list[CustomFunctionAnalyser]:
        return self._analysers

    @_analysers_by_fully_qualified_name.default
    def set_analysers_by_fully_qualified_name(self) -> None:
        return {analyser.qualified_name: analyser for analyser in self.analysers}

    def register_assertors(
        self,
        new_assertors: Iterable[Assertor],
    ) -> None:
        self._assertors = list(set(self._assertors) | set(new_assertors))

    def register_analysers(
        self,
        new_analysers: Iterable[CustomFunctionAnalyser],
    ) -> None:
        self._analysers = list(set(self._analysers) | set(new_analysers))
        self._analysers_by_fully_qualified_name = {
            analyser.qualified_name: analyser for analyser in self._analysers
        }

    def has_analyser(
        self,
        target: Symbol | None,
        *,
        modulename: ModuleName | None = None,
    ) -> bool:
        return self.get_analyser(target, modulename=modulename) is not None

    def get_analyser(
        self,
        target: Symbol | None,
        *,
        modulename: ModuleName | None = None,
    ) -> CustomFunctionAnalyser | None:
        if target is None:
            return None

        qualified_name = derive_qualified_name(target, modulename=modulename)

        if qualified_name in self._analysers_by_fully_qualified_name:
            return self._analysers_by_fully_qualified_name[qualified_name]

        return None


plugins = Plugins(assertors=[], analysers=DEFAULT_FUNCTION_ANALYSERS)


def register_rattr_plugins(
    *,
    assertors: Iterable[Assertor] = (),
    analysers: Iterable[CustomFunctionAnalyser] = (),
) -> None:
    """Configure the global rattr plugins.

    This should be called exactly once during custom rattr setup. If not called this
    remains default.
    """
    global plugins

    plugins.register_assertors(assertors)
    plugins.register_analysers(analysers)


def register_rattr_module_blacklist_patterns(*terms: ModuleName) -> None:
    """Register additional module blacklist terms."""
    config = Config()

    for term in terms:
        config.PLUGINS_BLACKLIST_PATTERNS.add(term)


def derive_qualified_name(
    symbol: Symbol,
    *,
    modulename: ModuleName | None = None,
) -> Identifier:
    if isinstance(symbol, Builtin):
        return symbol.id

    if isinstance(symbol, Import):
        return symbol.qualified_name

    if modulename is not None:
        return f"{modulename}.{symbol.id}"

    return symbol.id



"""Base classes for Rattr components."""
from __future__ import annotations

import ast
from abc import ABCMeta, abstractmethod, abstractproperty
from ast import NodeTransformer, NodeVisitor  # noqa: F401
from typing import TYPE_CHECKING

from rattr import error
from rattr.analyser.types import FunctionIr
from rattr.models.context import Context

if TYPE_CHECKING:
    from rattr.ast.types import Identifier


class Assertor(NodeVisitor):
    """Assertor base class.

    An assertor can either be strict or non-strict, set by the constructor
    argument `is_strict`, having the following affect:
        is_strict:
            on condition failure, log fatal error (halts execution)
        not is_strict:
            on condition failure, log warning

    """

    def __init__(self, is_strict: bool = True) -> None:
        self.is_strict: bool = is_strict
        self.context: Context = None

    def assert_holds(self, node: ast.AST, context: Context) -> None:
        """Entry point for an Assertor, visit the tree to assert properties."""
        self.context = context

        super().visit(node)

    def failed(self, message: str, culprit: ast.AST | None = None) -> None:
        """Handle assertion failure."""
        if self.is_strict:
            handler = error.fatal
        else:
            handler = error.warning

        handler(message, culprit)


class CustomFunctionAnalyser(NodeVisitor, metaclass=ABCMeta):
    """Base class for a custom function visitor."""

    @abstractproperty
    def name(self) -> Identifier:
        """Return the name of the function handled by this analyser."""
        return ""

    @abstractproperty
    def qualified_name(self) -> Identifier:
        """Return the qualified name of the function."""
        return ""

    @abstractmethod
    def on_def(
        self,
        name: Identifier,
        node: ast.FunctionDef | ast.AsyncFunctionDef,
        ctx: Context,
    ) -> FunctionIr:
        """Return the IR of the definition of the handled function."""
        return {
            "sets": set(),
            "gets": set(),
            "dels": set(),
            "calls": set(),
        }

    @abstractmethod
    def on_call(self, name: Identifier, node: ast.Call, ctx: Context) -> FunctionIr:
        """Return the IR produced by a call to the handled function.

        The returned IR will be union'd with the IR of the caller function.

        Argument `fna` is the FunctionAnalyser instance that the call is from.

        """
        return {
            "sets": set(),
            "gets": set(),
            "dels": set(),
            "calls": set(),
        }



from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from rattr.cli.toml import _load_from_file

if TYPE_CHECKING:
    from typing import Any


def find_data_file(name: str) -> Path:
    here = Path(__file__).resolve().parent

    data_file = here / "data" / name
    assert data_file.is_file()

    return data_file


@pytest.fixture
def illegal_field_name() -> str:
    return "illegal-field"


@pytest.fixture
def required_sys_args() -> list[str]:
    return ["my/rattr/target.py"]


@pytest.fixture
def required_sys_args_rattr_target() -> str:
    if sys.platform == "win32":
        return "my\\rattr\\target.py"
    return "my/rattr/target.py"


@pytest.fixture
def toml_well_formed_path() -> Path:
    return find_data_file("well_formed.toml")


@pytest.fixture
def toml_well_formed(toml_well_formed_path) -> dict[str, Any]:
    return _load_from_file(toml_well_formed_path).get("tool", {}).get("rattr", {})


@pytest.fixture
def toml_with_illegal_field(illegal_field_name, toml_well_formed) -> dict[str, Any]:
    return {**toml_well_formed, **{illegal_field_name: "any-old-value"}}


@pytest.fixture
def toml_override_path() -> Path:
    return find_data_file("override.toml")


@pytest.fixture
def toml_override(toml_override_path) -> dict[str, Any]:
    return _load_from_file(toml_override_path).get("tool", {}).get("rattr", {})



from __future__ import annotations

from collections.abc import MutableMapping
from typing import TYPE_CHECKING

import attrs
from attrs import field

from rattr.models.results.function import FunctionResults

if TYPE_CHECKING:
    from collections.abc import Iterator

    from rattr.versioning.typing import TypeAlias

FunctionName: TypeAlias = str


@attrs.mutable
class FileResults(MutableMapping[FunctionName, FunctionResults]):
    _function_results: dict[FunctionName, FunctionResults] = field(
        alias="function_results",
        factory=dict,
    )

    # ================================================================================ #
    # Mutable mapping abstract methods and mixin-overrides
    # ================================================================================ #

    def __getitem__(self, __key: FunctionName) -> FunctionResults:
        return self._function_results.__getitem__(__key)

    def __setitem__(
        self,
        __key: FunctionName,
        __value: FunctionResults,
    ) -> None:
        return self._function_results.__setitem__(__key, __value)

    def __delitem__(self, __key: FunctionName) -> None:
        return self._function_results.__delitem__(__key)

    def __iter__(self) -> Iterator[FunctionName]:
        return self._function_results.__iter__()

    def __len__(self) -> int:
        return self._function_results.__len__()

    def clear(self) -> None:
        # NOTE Better than derived implementation
        return self._function_results.clear()



from __future__ import annotations

from pathlib import Path
from unittest import mock

from rattr.config._util import _is_project_root
from rattr.config.util import find_xdg_cache_dir


def mock_path(
    loc: str,
    *,
    is_dir: bool = False,
    is_file: bool = False,
):
    m_path = mock.Mock(spec=Path)

    m_path.__str__ = lambda _self: loc

    m_path.is_dir.return_value = is_dir
    m_path.is_file.return_value = is_file
    m_path.exists.return_value = is_dir or is_file

    return m_path


class TestIsProjectRoot:
    def test_is_not_dir(self):
        # Definitionally a root directory must be a directory!
        assert not _is_project_root(mock_path("/some/file.txt", is_dir=False))


class TestFindXdgCacheDir:
    @mock.patch("rattr.config.util.os")
    @mock.patch("rattr.config.util.Path")
    def test_user_xdg_cache_home_is_set(self, m_path, m_os):
        xdg_cache_home = "/home/user/.my_custom_cache"

        m_path.side_effect = lambda loc: mock_path(loc, is_dir=True)
        m_os.environ = {"XDG_CACHE_HOME": xdg_cache_home}

        assert str(find_xdg_cache_dir()) == xdg_cache_home

    @mock.patch("rattr.config.util.os")
    def test_user_xdg_cache_home_is_not_set(self, m_os):
        m_os.environ = {}
        assert str(find_xdg_cache_dir()) == str(Path.home() / ".cache")



from __future__ import annotations

from typing import TYPE_CHECKING

from rattr.analyser.file import FileAnalyser
from rattr.models.context import Context, compile_root_context
from rattr.models.ir import FileIr
from rattr.models.symbol import Call, CallArguments, CallInterface, Func, Name

if TYPE_CHECKING:
    from tests.shared import MakeSymbolTableFn, ParseFn, ParseWithContextFn


class TestCustomFunctionAnalysers:
    def test_sorted_no_key(self, parse: ParseFn, make_symbol_table: MakeSymbolTableFn):
        _ast = parse(
            """
            def a_func(arg):
                sorted(arg)
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table([a_func], include_root_symbols=True),
            ),
            file_ir={
                a_func: {
                    "gets": {
                        Name("arg"),
                    },
                    "sets": set(),
                    "dels": set(),
                    "calls": set(),
                }
            },
        )

        assert results == expected

    def test_sorted_with_constant_key(
        self,
        parse: ParseFn,
        make_symbol_table: MakeSymbolTableFn,
    ):
        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))

        # Literal
        _ast = parse(
            """
            def a_func(arg):
                sorted(arg, key=1)
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table([a_func], include_root_symbols=True),
            ),
            file_ir={
                a_func: {
                    "gets": {
                        Name("arg"),
                    },
                    "sets": set(),
                    "dels": set(),
                    "calls": set(),
                }
            },
        )

        assert results == expected

        # Attribute
        _ast = parse(
            """
            def a_func(arg):
                sorted(arg, key=arg.attr)
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table([a_func], include_root_symbols=True),
            ),
            file_ir={
                a_func: {
                    "gets": {
                        Name("arg"),
                        Name("arg.attr", "arg"),
                    },
                    "sets": set(),
                    "dels": set(),
                    "calls": set(),
                }
            },
        )

        assert results == expected

    def test_sorted_key_is_callable(
        self,
        parse: ParseFn,
        make_symbol_table: MakeSymbolTableFn,
    ):
        # Via lambda
        _ast = parse(
            """
            def a_func(arg):
                sorted(arg, key=lambda a: key_func(a))

            def key_func(a):
                return a.b
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))
        key_func = Func(name="key_func", interface=CallInterface(args=("a",)))
        key_func_call = Call(
            name="key_func",
            args=CallArguments(args=("a",), kwargs={}),
            target=key_func,
        )

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table(
                    [
                        a_func,
                        key_func,
                    ],
                    include_root_symbols=True,
                ),
            ),
            file_ir={
                a_func: {
                    "gets": {
                        Name("arg"),
                    },
                    "sets": set(),
                    "dels": set(),
                    "calls": {key_func_call},
                },
                key_func: {
                    "gets": {
                        Name("a.b", "a"),
                    },
                    "sets": set(),
                    "dels": set(),
                    "calls": set(),
                },
            },
        )

        assert results == expected

    def test_sorted_key_is_attribute_or_element(
        self,
        parse: ParseFn,
        make_symbol_table: MakeSymbolTableFn,
    ):
        a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))

        # Attribute
        _ast = parse(
            """
            def a_func(arg):
                sorted(arg, key=lambda a: a.attr)
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table([a_func], include_root_symbols=True),
            ),
            file_ir={
                a_func: {
                    "gets": {
                        Name("arg"),
                        Name("arg.attr", "arg"),
                    },
                    "sets": set(),
                    "dels": set(),
                    "calls": set(),
                }
            },
        )

        assert results == expected

        # Element
        _ast = parse(
            """
            def a_func(arg):
                sorted(arg, key=lambda a: a[0])
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table([a_func], include_root_symbols=True),
            ),
            file_ir={
                a_func: {
                    "gets": {
                        Name("arg"),
                        Name("arg[]", "arg"),
                    },
                    "sets": set(),
                    "dels": set(),
                    "calls": set(),
                }
            },
        )

        assert results == expected

        # Complex
        _ast = parse(
            """
            def a_func(arg):
                sorted(arg, key=lambda a: a[0].attr)
            """
        )
        results = FileAnalyser(_ast, compile_root_context(_ast)).analyse()

        expected = FileIr(
            context=Context(
                parent=None,
                symbol_table=make_symbol_table([a_func], include_root_symbols=True),
            ),
            file_ir={
                a_func: {
                    "gets": {
                        Name("arg"),
                        Name("arg[].attr", "arg"),
                    },
                    "sets": set(),
                    "dels": set(),
                    "calls": set(),
                }
            },
        )

        assert results == expected


def test_getattr(
    parse_with_context: ParseWithContextFn,
    make_symbol_table: MakeSymbolTableFn,
):
    ast_module, context = parse_with_context(
        """
        def a_func(arg):
            return getattr(getattr(getattr(arg, "inner_attr"), "middle_attr"), "outer_attr")
        """
    )
    results = FileAnalyser(ast_module, context).analyse()

    a_func = Func(name="a_func", interface=CallInterface(args=("arg",)))

    expected = FileIr(
        context=Context(
            parent=None,
            symbol_table=make_symbol_table([a_func], include_root_symbols=True),
        ),
        file_ir={
            a_func: {
                "gets": {
                    Name("arg"),
                    Name("arg.inner_attr", "arg"),
                    Name("arg.inner_attr.middle_attr", "arg"),
                    Name("arg.inner_attr.middle_attr.outer_attr", "arg"),
                },
                "sets": set(),
                "dels": set(),
                "calls": set(),
            }
        },
    )

    assert results == expected



"""Private util functions."""
from __future__ import annotations

import ast
from typing import TYPE_CHECKING

from rattr import error

if TYPE_CHECKING:
    from typing import Final, NoReturn


_feature_not_supported: Final = "{feature} not supported in function calls"

PYTHON_BUILTINS_LOCATION: Final = "built-in"
"""Python's built-in function location.

This is the module spec origin for several stdlib modules.
"""


def arg_name(arg: ast.expr, *, culprit: ast.Call | None = None) -> str:
    """Return the safely-evaluated argument name."""
    from rattr.analyser.util import get_fullname  # circular dep.

    if isinstance(arg, ast.Starred):
        error.error(
            _feature_not_supported.format(feature="iterable unpacking"),
            culprit=culprit,
        )

    return get_fullname(arg, safe=True)


def kwarg_name(
    kwarg: ast.keyword,
    *,
    culprit: ast.Call | None = None,
) -> str | NoReturn:
    """Return the safely-evaluated keyword-argument name."""
    from rattr.analyser.util import get_fullname  # circular dep.

    if kwarg.arg is None:
        error.fatal(
            _feature_not_supported.format(feature="dictionary unpacking"),
            culprit=culprit,
        )

    return get_fullname(kwarg.value, safe=True)



from __future__ import annotations

from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING

from rattr.config._types import Config

if TYPE_CHECKING:
    from collections.abc import Generator


@contextmanager
def enter_file(new_file: Path | str | None) -> Generator[None, None, None]:
    """Set the config state's current file to the given file while in scope.

    >>> config = Config()
    >>> visited: list[Path] = []
    >>> with enter_file(Path("target_file.py")):
    ...     visited.append(config.state.current_file)
    ...     with enter_file(Path("an_imported_file.py")):
    ...         visited.append(config.state.current_file)
    ...     visited.append(config.state.current_file)
    >>> print(visited)
    [Path("target_file.py"), Path("an_imported_file.py"), Path("target_file.py")]
    """
    config = Config()

    if isinstance(new_file, str):
        new_file = Path(new_file)

    old_file = config.state.current_file
    config.state.current_file = new_file

    yield

    config.state.current_file = old_file



from __future__ import annotations

import re
import sys
from functools import cache
from importlib.util import find_spec as stdlib_find_spec
from pathlib import Path
from typing import TYPE_CHECKING

from isort import sections
from isort.api import place_module

from rattr.config import Config
from rattr.module_locator._locate import (  # noqa: F401
    find_module_in_path,
    iter_python_path_dirs,
    locate_module_in_python_path,
)
from rattr.module_locator.models import ModuleSpec

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from typing import Final

    from rattr.ast.types import FullyQualifiedName, ModuleName
    from rattr.versioning.typing import TypeAlias

    ImportLevel: TypeAlias = int


RE_PIP_INSTALL_LOCATIONS: Final = (re.compile(r".+/site-packages.*"),)


def module_exists(modulename: ModuleName) -> bool:
    return find_module_spec_fast(modulename) is not None


@cache
def find_module_spec_fast(modulename: ModuleName) -> ModuleSpec | None:
    if is_in_stdlib(modulename):
        return __find_stdlib_module_spec_impl(modulename)

    locations = locate_module_in_python_path(modulename)

    if not locations:
        return None

    return ModuleSpec(name=modulename, origin=format_origin_for_os(locations[0]))


@cache
def __find_stdlib_module_spec_impl(modulename: ModuleName) -> ModuleSpec | None:
    try:
        stdlib_spec = stdlib_find_spec(modulename)
    except (AttributeError, ModuleNotFoundError, ValueError):
        return None

    if stdlib_spec is None:
        return None

    return ModuleSpec(
        name=stdlib_spec.name,
        origin=format_origin_for_os(stdlib_spec.origin),
    )


@cache
def format_origin_for_os(origin: str | Path | None) -> str | None:
    if origin is None:
        return origin

    if isinstance(origin, Path):
        origin = str(origin)

    if sys.platform != "win32":
        return origin

    # HACK
    # Windows paths are case-insensitive so force lower case for comparisons etc
    return origin[:2] + origin[2:].lower()


@cache
def find_module_name_and_spec(
    target: FullyQualifiedName,
) -> tuple[ModuleName, ModuleSpec] | tuple[None, None]:
    """Return the module name and spec for the given fully qualified name.

    The fully qualified name is that of the module:
    >>> find_module_name_and_spec("math")
    ("math", ModuleSpec(name='math', loader=... origin='built-in')

    The fully qualified name is that of a module member:
    >>> find_module_name_and_spec("math.pi")
    ("math", ModuleSpec(name='math', loader=... origin='built-in')

    The fully qualified name is that of a member of a nested module:
    >>> find_module_name_and_spec("os.path.join")
    ("os.path", ModuleSpec(name='posixpath', loader=..., origin=...))

    The given name is not a fully qualified name:
    >>> find_module_name_and_spec("join")
    (None, None)
    """
    # Can't so anything with relative imports
    if target.startswith("."):
        return None, None

    for modulename in iter_module_names_right(target.split(".")):
        if (spec := find_module_spec_fast(modulename)) is not None:
            return modulename, spec

    return None, None


def iter_module_names_right(module_parts: Iterable[str]) -> Iterator[ModuleName]:
    """Return the possible module qualified names for the given parts.

    >>> list(iter_possible_modules([]))
    []

    >>> list(iter_possible_modules(["module"]))
    ["module"]

    >>> list(iter_possible_modules(["module", "sub_module"]))
    ['module.sub_module', 'module']

    >>> list(iter_possible_modules(["module", "sub_module", "file"]))
    ['module.sub_module.file', 'module.sub_module', 'module']
    """
    yield ".".join(module_parts)
    for end_offset in range(1, len(module_parts)):
        yield ".".join(module_parts[:-end_offset])


def derive_module_names_right(modulename: ModuleName) -> list[ModuleName]:
    return list(iter_module_names_right(modulename.split(".")))


@cache
def derive_module_name_from_path(filepath: Path | str | None) -> str | None:
    """Return the recognised name of the given module."""
    if filepath is None:
        raise ValueError("unable to derive the module of 'None'")

    filepath = Path(filepath)
    filepath_str = str(filepath)

    # Given a file '/path/to/repo/my_module/file/__init.py', the longest possible module
    # name would be 'path.to.repo.my_module.file' though in reality it may just be
    # 'my_module.file'
    longest_possible_modulename = (
        filepath_str.replace("/", ".")
        .replace("\\", ".")
        .removesuffix(".__init__.py")
        .removesuffix(".py")
        .strip(".")  # strip remaining "." in relative files
    )

    for module in iter_module_names_left(longest_possible_modulename.split(".")):
        if module_exists(module):
            return module

    return None


def iter_module_names_left(module_parts: Iterable[str]) -> Iterator[ModuleName]:
    """Return the possible module qualified names for the given parts.

    >>> list(iter_possible_modules([]))
    []

    >>> list(iter_possible_modules(["module"]))
    ["module"]

    >>> list(iter_possible_modules(["module", "sub_module"]))
    ['module.sub_module', 'sub_module']

    >>> list(iter_possible_modules(["module", "sub_module", "file"]))
    ['module.sub_module.file', 'sub_module.file', 'file']
    """
    for start_offset in range(len(module_parts)):
        yield ".".join(module_parts[start_offset:])


@cache
def derive_module_names_left(modulename: ModuleName) -> list[ModuleName]:
    return list(iter_module_names_left(modulename.split(".")))


@cache
def derive_absolute_module_name(
    base: ModuleName,
    target: ModuleName | None,
    level: ImportLevel,
) -> ModuleName:
    config = Config()

    if config.state.current_file.name == "__init__.py":
        level -= 1

    if level > 0:
        base = ".".join(base.split(".")[:-level])

    return f"{base}.{target}" if target is not None else base


@cache
def is_in_stdlib(name: ModuleName) -> bool:
    """Return `True` if the given name is an stdlib module or in an stdlib module.

    >>> is_stdlib_module("math")
    True
    >>> is_stdlib_module("math.pi")
    True
    >>> is_stdlib_module("pytest.fixture")
    False
    """
    return place_module(name) == sections.STDLIB


@cache
def is_in_pip(name: ModuleName) -> bool:
    """Return `True` if the given name is available via pip.

    >>> # Given that pytest is pip installed
    >>> is_in_pip("pytest")
    True
    >>> is_in_pip("pytest.fixture")
    True
    >>> is_in_pip("math")
    False
    >>> is_in_pip("something.made.up")
    False
    """
    return any(
        re_pip_location.fullmatch(origin)
        for module in derive_module_names_right(name)
        for re_pip_location in RE_PIP_INSTALL_LOCATIONS
        if (origin := __safe_origin(module)) is not None
    )


@cache
def is_in_import_blacklist(name: ModuleName) -> bool:
    """Return `True` if the given name matches a blacklisted pattern."""
    config = Config()

    # Somewhat undefined behaviour, but don't follow null strings
    if not name:
        return True

    # Exclude stdlib modules such as the built-in "_thread" (stdlib modules are handled
    # separately from blacklist).
    if is_in_stdlib(name):
        return False

    origins = [__safe_origin(module) for module in derive_module_names_right(name)]

    # It is possible that we can't determine the spec (due to some PYTHON_PATH
    # tampering or some other shenanigans) and in that case we still want to assume that
    # the origin is exactly as given as it may still be blacklisted.
    origins.append(name)

    return any(
        re_pattern.fullmatch(origin)
        for origin in origins
        for re_pattern in config.re_blacklist_patterns
        if origin is not None
    )


@cache
def __safe_origin(module: ModuleName) -> str | None:
    spec = find_module_spec_fast(module)

    if spec is None or spec.origin is None:
        return None

    # No backslashes, bad windows!
    return spec.origin.replace("\\", "/")



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from rattr.models.plugins import Plugins
from rattr.models.symbol import Builtin, CallInterface, Func, Import

if TYPE_CHECKING:
    from rattr.analyser.base import Assertor, CustomFunctionAnalyser


@pytest.fixture
def example_plugins(
    example_assertor: Assertor,
    builtin_print_analyser: CustomFunctionAnalyser,
    example_func_analyser: CustomFunctionAnalyser,
):
    return Plugins(
        assertors=[example_assertor],
        analysers=[builtin_print_analyser, example_func_analyser],
    )


def test_register_assertors(example_assertor: Assertor):
    plugins = Plugins()
    assert plugins.assertors == []

    plugins.register_assertors((example_assertor, example_assertor, example_assertor))
    assert plugins.assertors == [example_assertor]


def test_register_analysers(
    builtin_print_analyser: CustomFunctionAnalyser,
    example_func_analyser: CustomFunctionAnalyser,
):
    plugins = Plugins()
    assert plugins.analysers == []

    plugins.register_analysers(
        (
            builtin_print_analyser,
            example_func_analyser,
            example_func_analyser,
        )
    )
    assert plugins.analysers == [example_func_analyser]


def test_plugins_has_analyser(example_plugins: Plugins):
    assert example_plugins.has_analyser(Builtin("print"))
    assert not example_plugins.has_analyser(Builtin("getattr"))


@pytest.mark.parametrize("modulename", [None, "module", "bleep.bloop"])
def test_plugins_get_analyser_builtin(
    example_plugins: Plugins,
    builtin_print_analyser: CustomFunctionAnalyser,
    modulename: str | None,
):
    assert (
        example_plugins.get_analyser(Builtin("print"), modulename=modulename)
        == builtin_print_analyser
    )
    assert (
        example_plugins.get_analyser(Builtin("not_print"), modulename=modulename)
        is None
    )


def test_plugins_get_analyser_local_function_positive(
    example_plugins: Plugins,
    example_func_analyser: CustomFunctionAnalyser,
):
    assert (
        example_plugins.get_analyser(
            Func("example", interface=CallInterface()),
            modulename=None,
        )
        == example_func_analyser
    )
    assert (
        example_plugins.get_analyser(
            Func("example", interface=CallInterface()),
            modulename="module",
        )
        == example_func_analyser
    )
    assert (
        example_plugins.get_analyser(
            Func("example", interface=CallInterface()),
            modulename="not_the_module_example.is_defined_in",
        )
        is None
    )


@pytest.mark.parametrize("modulename", [None, "module", "bleep.bloop"])
def test_plugins_get_analyser_local_function_negative(
    example_plugins: Plugins,
    modulename: str | None,
):
    assert (
        example_plugins.get_analyser(
            Func("nope", interface=CallInterface()),
            modulename=modulename,
        )
        is None
    )


@pytest.mark.parametrize("modulename", [None, "module", "bleep.bloop"])
def test_plugins_get_analyser_import_positive(
    example_plugins: Plugins,
    example_func_analyser: CustomFunctionAnalyser,
    modulename: str | None,
):
    assert (
        example_plugins.get_analyser(
            Import(name="example", qualified_name="different.module.example"),
            modulename=modulename,
        )
        is None
    )
    assert (
        example_plugins.get_analyser(
            Import(name="example", qualified_name="module.example"),
            modulename=modulename,
        )
        == example_func_analyser
    )


@pytest.mark.parametrize("modulename", [None, "module", "bleep.bloop"])
def test_plugins_get_analyser_import_negative(
    example_plugins: Plugins,
    modulename: str | None,
):
    assert (
        example_plugins.get_analyser(
            Import(name="no_analyser", qualified_name="different.module.no_analyser"),
            modulename=modulename,
        )
        is None
    )
    assert (
        example_plugins.get_analyser(
            Import(name="no_analyser", qualified_name="module.no_analyser"),
            modulename=modulename,
        )
        is None
    )
